"""Visualizer for results of anomaly task prediction."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Optional

import cv2
import numpy as np

from otx.api.entities.annotation import AnnotationSceneEntity

from .visualizer import Visualizer


class AnomalyVisualizer(Visualizer):
    """Visualize the predicted output by drawing the annotations on the input image.

    Example:
        >>> predictions = inference_model.predict(frame)
        >>> annotation = prediction_converter.convert_to_annotation(predictions)
        >>> output = visualizer.draw(frame, annotation.shape, annotation.get_labels())
        >>> visualizer.show(output)
    """

    def __init__(
        self,
        window_name: Optional[str] = None,
        show_count: bool = False,
        is_one_label: bool = False,
        no_show: bool = False,
        delay: Optional[int] = None,
    ) -> None:
        super().__init__(window_name, show_count, is_one_label, no_show, delay)
        if not no_show:
            cv2.namedWindow(
                self.window_name,
                cv2.WINDOW_NORMAL | cv2.WINDOW_KEEPRATIO | cv2.WINDOW_GUI_EXPANDED,
            )
            self.trackbar_name = "Opacity"
            cv2.createTrackbar(self.trackbar_name, self.window_name, 0, 100, lambda x: x)

    @staticmethod
    def to_heat_mask(mask: np.ndarray) -> np.ndarray:
        """Create heat mask from saliency map.

        Args:
            mask: saliency map
        """
        heat_mask = cv2.normalize(mask, None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX).astype(np.uint8)
        return cv2.applyColorMap(heat_mask.astype(np.uint8), cv2.COLORMAP_JET)

    # pylint:disable=signature-differs
    def draw(  # type: ignore[override]
        self, image: np.ndarray, annotation: AnnotationSceneEntity, meta: dict
    ) -> np.ndarray:
        """Draw annotations on the image.

        Args:
            image: Input image
            annotation: Annotations to be drawn on the input image
            metadata: Metadata with saliency map

        Returns:
            Output image with annotations.
        """

        heat_mask = self.to_heat_mask(1 - meta["anomaly_map"])
        alpha = cv2.getTrackbarPos(self.trackbar_name, self.window_name) / 100.0
        image = (1 - alpha) * image + alpha * heat_mask
        image = cv2.cvtColor(image.astype(np.uint8), cv2.COLOR_RGB2BGR)

        return self.shape_drawer.draw(image, annotation, labels=[])



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from collections import defaultdict

import pytest
from mmcv import Config

from otx.algorithms.segmentation.adapters.mmseg.utils.config_utils import (
    patch_adaptive_repeat_dataset,
    patch_config,
    patch_datasets,
    prepare_for_training,
    rescale_num_iterations,
    set_data_classes,
    set_distributed_mode,
    set_hyperparams,
    set_num_classes,
)
from otx.algorithms.segmentation.configs.base import SegmentationConfig
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import Domain, LabelEntity
from otx.api.usecases.reporting.time_monitor_callback import TimeMonitorCallback
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)


class TestConfigUtilsInputParamsValidation:
    @e2e_pytest_unit
    def test_patch_config_input_params_validation(self):
        """
        <b>Description:</b>
        Check "patch_config" function input parameters validation

        <b>Input data:</b>
        "patch_config" function input parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "patch_config" function
        """
        label = LabelEntity(name="test label", domain=Domain.SEGMENTATION)
        correct_values_dict = {
            "config": Config(),
            "work_dir": "./work_dir",
            "labels": [label],
        }
        unexpected_float = 1.1
        unexpected_values = [
            # Unexpected float is specified as "config" parameter
            ("config", unexpected_float),
            # Unexpected float is specified as "work_dir" parameter
            ("work_dir", unexpected_float),
            # Empty string is specified as "work_dir" parameter
            ("work_dir", ""),
            # String with null-character is specified as "work_dir" parameter
            ("work_dir", "null\0character/path"),
            # String with non-printable character is specified as "work_dir" parameter
            ("work_dir", "\non_printable_character/path"),
            # Unexpected float is specified as "labels" parameter
            ("labels", unexpected_float),
            # Unexpected float is specified as nested "label"
            ("labels", [label, unexpected_float]),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=patch_config,
        )

    @e2e_pytest_unit
    def test_set_hyperparams_input_params_validation(self):
        """
        <b>Description:</b>
        Check "set_hyperparams" function input parameters validation

        <b>Input data:</b>
        "set_hyperparams" function input parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "set_hyperparams" function
        """
        correct_values_dict = {
            "config": Config(),
            "hyperparams": SegmentationConfig(),
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "config" parameter
            ("config", unexpected_str),
            # Unexpected string is specified as "hyperparams" parameter
            ("hyperparams", unexpected_str),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=set_hyperparams,
        )

    @e2e_pytest_unit
    def test_rescale_num_iterations_input_params_validation(self):
        """
        <b>Description:</b>
        Check "rescale_num_iterations" function input parameters validation

        <b>Input data:</b>
        "rescale_num_iterations" function input parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "rescale_num_iterations" function
        """
        correct_values_dict = {
            "config": Config(),
            "schedule_scale": 1.1,
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "config" parameter
            ("config", unexpected_str),
            # Unexpected string is specified as "schedule_scale" parameter
            ("schedule_scale", unexpected_str),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=rescale_num_iterations,
        )

    @e2e_pytest_unit
    def test_patch_adaptive_repeat_dataset_input_params_validation(self):
        """
        <b>Description:</b>
        Check "patch_adaptive_repeat_dataset" function input parameters validation

        <b>Input data:</b>
        "patch_adaptive_repeat_dataset" function input parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "patch_adaptive_repeat_dataset" function
        """
        correct_values_dict = {"config": Config(), "num_samples": 10}
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "config" parameter
            ("config", unexpected_str),
            # Unexpected string is specified as "num_samples" parameter
            ("num_samples", unexpected_str),
            # Unexpected string is specified as "decay" parameter
            ("decay", unexpected_str),
            # Unexpected string is specified as "factor" parameter
            ("factor", unexpected_str),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=patch_adaptive_repeat_dataset,
        )

    @e2e_pytest_unit
    def test_prepare_for_training_input_params_validation(self):
        """
        <b>Description:</b>
        Check "prepare_for_training" function input parameters validation

        <b>Input data:</b>
        "prepare_for_training" function input parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "prepare_for_training" function
        """
        dataset = DatasetEntity()
        time_monitor = TimeMonitorCallback(num_epoch=5, num_train_steps=2, num_val_steps=1, num_test_steps=1)
        correct_values_dict = {
            "config": Config(),
            "train_dataset": dataset,
            "val_dataset": dataset,
            "time_monitor": time_monitor,
            "learning_curves": defaultdict(),
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "config" parameter
            ("config", unexpected_str),
            # Unexpected string is specified as "train_dataset" parameter
            ("train_dataset", unexpected_str),
            # Unexpected string is specified as "val_dataset" parameter
            ("val_dataset", unexpected_str),
            # Unexpected string is specified as "time_monitor" parameter
            ("time_monitor", unexpected_str),
            # Unexpected string is specified as "learning_curves" parameter
            ("learning_curves", unexpected_str),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=prepare_for_training,
        )

    @e2e_pytest_unit
    def test_set_distributed_mode_input_params_validation(self):
        """
        <b>Description:</b>
        Check "set_distributed_mode" function input parameters validation

        <b>Input data:</b>
        "set_distributed_mode" function input parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "set_distributed_mode" function
        """
        correct_values_dict = {"config": Config(), "distributed": True}
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "config" parameter
            ("config", unexpected_str),
            # Unexpected string is specified as "distributed" parameter
            ("distributed", unexpected_str),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=set_distributed_mode,
        )

    @e2e_pytest_unit
    def test_set_data_classes_input_params_validation(self):
        """
        <b>Description:</b>
        Check "set_data_classes" function input parameters validation

        <b>Input data:</b>
        "set_data_classes" function input parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "set_data_classes" function
        """
        label_name = "label_1"
        correct_values_dict = {"config": Config(), "label_names": [label_name]}
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "config" parameter
            ("config", unexpected_int),
            # Unexpected integer is specified as "label_names" parameter
            ("label_names", unexpected_int),
            # Unexpected integer is specified as nested "label"
            ("label_names", [label_name, unexpected_int]),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=set_data_classes,
        )

    @e2e_pytest_unit
    def test_set_num_classes_input_params_validation(self):
        """
        <b>Description:</b>
        Check "set_num_classes" function input parameters validation

        <b>Input data:</b>
        "set_num_classes" function input parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "set_num_classes" function
        """
        correct_values_dict = {"config": Config(), "num_classes": 1}
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "config" parameter
            ("config", unexpected_str),
            # Unexpected string is specified as "num_classes" parameter
            ("num_classes", unexpected_str),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=set_num_classes,
        )

    @e2e_pytest_unit
    def test_patch_datasets_input_params_validation(self):
        """
        <b>Description:</b>
        Check "patch_datasets" function input parameters validation

        <b>Input data:</b>
        "config" non-Config type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "patch_datasets" function
        """
        with pytest.raises(ValueError):
            patch_datasets(config="unexpected string")  # type: ignore



"""Test for otx.mpa.modules.optimizer.lars."""
import pytest
import torch
import torch.nn as nn

from otx.mpa.modules.optimizer.lars import LARS
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.mpa.test_helpers import (
    generate_random_torch_image,
    generate_toy_cnn_model,
)


class TestLARS:
    def setup(self):
        self.model = generate_toy_cnn_model()
        self.model.train()

        self.loss = nn.MSELoss()

    @e2e_pytest_unit
    def test_init(self):
        negative_lr = -0.1
        with pytest.raises(ValueError, match="Invalid learning rate: {}".format(negative_lr)):
            LARS(self.model.parameters(), lr=negative_lr)

        lr = 1e-1
        negative_momentum = -0.1
        with pytest.raises(ValueError, match="Invalid momentum value: {}".format(negative_momentum)):
            LARS(self.model.parameters(), lr=lr, momentum=negative_momentum)

        negative_wd = -0.1
        with pytest.raises(ValueError, match="Invalid weight_decay value: {}".format(negative_wd)):
            LARS(self.model.parameters(), lr=lr, weight_decay=negative_wd)

        negative_eta = -0.1
        with pytest.raises(ValueError, match="Invalid LARS coefficient value: {}".format(negative_eta)):
            LARS(self.model.parameters(), lr=lr, eta=negative_eta)

        nesterov = True
        with pytest.raises(ValueError, match="Nesterov momentum requires a momentum and zero dampening"):
            LARS(self.model.parameters(), lr=lr, nesterov=nesterov)

        LARS(self.model.parameters(), lr=lr, exclude_bn_from_weight_decay=True)

    @e2e_pytest_unit
    def test_step(self):
        selfsl_optimizer = LARS(self.model.parameters(), lr=0.1, mode="selfsl", exclude_bn_from_weight_decay=True)
        selfsl_optimizer.zero_grad()
        input_img = generate_random_torch_image(1, 3, 3, 3)
        input_img.requires_grad = True
        target = torch.randn(1, 3)

        logit = self.model(input_img)

        loss = self.loss(logit.view(1, -1), target)
        loss.backward()
        selfsl_optimizer.step()

        optimizer = LARS(self.model.parameters(), lr=0.1, momentum=0.9, exclude_bn_from_weight_decay=True)
        optimizer.zero_grad()
        input_img = generate_random_torch_image(1, 3, 3, 3)
        input_img.requires_grad = True
        target = torch.randn(1, 3)

        logit = self.model(input_img)

        loss = self.loss(logit.view(1, -1), target)
        loss.backward()
        optimizer.step()



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
from mmcv.runner import OPTIMIZERS
from torch.optim.optimizer import Optimizer, required


@OPTIMIZERS.register_module()
class LARS(Optimizer):
    r"""Implements layer-wise adaptive rate scaling for SGD.

    Args:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float): base learning rate (\gamma_0)
        momentum (float, optional): momentum factor (default: 0) ("m")
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
            ("\beta")
        dampening (float, optional): dampening for momentum (default: 0)
        eta (float, optional): LARS coefficient
        nesterov (bool, optional): enables Nesterov momentum (default: False)

    Based on Algorithm 1 of the following paper by You, Gitman, and Ginsburg.
    Large Batch Training of Convolutional Networks:
        https://arxiv.org/abs/1708.03888

    Example:
        >>> optimizer = LARS(model.parameters(), lr=0.1, momentum=0.9,
        >>>                  weight_decay=1e-4, eta=1e-3)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()
    """

    def __init__(
        self,
        params,
        lr=required,
        momentum=0,
        dampening=0,
        weight_decay=0,
        eta=0.001,
        nesterov=False,
        mode=None,
        exclude_bn_from_weight_decay=False,
    ):
        if lr is not required and lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if momentum < 0.0:
            raise ValueError("Invalid momentum value: {}".format(momentum))
        if weight_decay < 0.0:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        if eta < 0.0:
            raise ValueError("Invalid LARS coefficient value: {}".format(eta))

        defaults = dict(
            lr=lr, momentum=momentum, dampening=dampening, weight_decay=weight_decay, nesterov=nesterov, eta=eta
        )
        if nesterov and (momentum <= 0 or dampening != 0):
            raise ValueError("Nesterov momentum requires a momentum and zero dampening")

        # split param group into weight_decay/non-weight decay.
        if exclude_bn_from_weight_decay:
            param_groups = list(params)
            if not isinstance(param_groups[0], dict):
                param_groups = [{"params": param_groups}]

            new_param_groups = []
            for param_group in param_groups:
                decay, no_decay = [], []
                for param in param_group.pop("params", []):
                    if not param.requires_grad:
                        continue

                    if len(param.shape) == 1:
                        no_decay.append(param)
                    else:
                        decay.append(param)

                decay_param_group = param_group.copy()
                decay_param_group["params"] = decay

                no_decay_param_group = param_group.copy()
                no_decay_param_group["params"] = no_decay
                no_decay_param_group["weight_decay"] = 0
                no_decay_param_group["lars_exclude"] = True

                new_param_groups.append(decay_param_group)
                new_param_groups.append(no_decay_param_group)

        self.mode = mode

        super(LARS, self).__init__(new_param_groups, defaults)

    def __setstate__(self, state):
        super(LARS, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault("nesterov", False)

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step.

        Args:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            weight_decay = group["weight_decay"]
            momentum = group["momentum"]
            dampening = group["dampening"]
            nesterov = group["nesterov"]
            eta = group["eta"]
            lars_exclude = group.get("lars_exclude", False)

            for p in group["params"]:
                if p.grad is None:
                    continue
                d_p = p.grad

                # Add weight decay before computing adaptive LR.
                # Seems to be pretty important in SIMclr style models.
                local_lr = 1.0
                if self.mode == "selfsl":
                    if weight_decay != 0:
                        d_p = d_p.add(p, alpha=weight_decay)
                    if not lars_exclude:
                        weight_norm = torch.norm(p).item()
                        grad_norm = torch.norm(d_p).item()
                        if weight_norm > 0 and grad_norm > 0:
                            local_lr = eta * weight_norm / grad_norm
                else:
                    if not lars_exclude:
                        weight_norm = torch.norm(p).item()
                        grad_norm = torch.norm(d_p).item()
                        local_lr = eta * weight_norm / (grad_norm + weight_decay * weight_norm)
                    if weight_decay != 0:
                        d_p = d_p.add(p, alpha=weight_decay)

                d_p = d_p.mul(local_lr)

                if momentum != 0:
                    param_state = self.state[p]
                    if "momentum_buffer" not in param_state:
                        buf = param_state["momentum_buffer"] = torch.clone(d_p).detach()
                    else:
                        buf = param_state["momentum_buffer"]
                        buf.mul_(momentum).add_(d_p, alpha=1 - dampening)
                    if nesterov:
                        d_p = d_p.add(buf, alpha=momentum)
                    else:
                        d_p = buf

                p.add_(d_p, alpha=-group["lr"])

        return loss



"""This module contains the base class to define ConfigurableParameters within OTX."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
from attr import attrib, attrs, setters

from otx.api.entities.id import ID

from .elements.parameter_group import ParameterGroup
from .elements.utils import convert_string_to_id
from .enums.config_element_type import ConfigElementType


@attrs(auto_attribs=True, order=False, eq=False)
class ConfigurableParameters(ParameterGroup):
    """Base class representing a generic set of configurable parameters.

    A ConfigurableParameters instance is essentially a parameter group with an id
    attached to it, so that it can be uniquely identified in the repositories.

    Attributes:
        id (ID): ID that uniquely identifies the ConfigurableParameters
        type (ConfigElementType): Type of the ConfigurableParameters
    """

    id: ID = attrib(default=ID(), kw_only=True, converter=convert_string_to_id)
    type: ConfigElementType = attrib(
        default=ConfigElementType.CONFIGURABLE_PARAMETERS,
        repr=False,
        init=False,
        on_setattr=setters.frozen,
    )



"""Patch mmcv and mpa stuff."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from copy import deepcopy

from otx.algorithms.common.adapters.nncf import (
    NNCF_PATCHER,
    is_nncf_enabled,
    no_nncf_trace_wrapper,
)

if is_nncf_enabled():
    from nncf.torch.nncf_network import NNCFNetwork

    # pylint: disable-next=ungrouped-imports
    from otx.algorithms.common.adapters.nncf.patches import nncf_train_step

    # add wrapper train_step method
    NNCFNetwork.train_step = nncf_train_step


# pylint: disable-next=unused-argument,invalid-name
def _evaluation_wrapper(self, fn, runner, *args, **kwargs):
    # TODO: move this patch to upper level (mmcv)
    # as this is not only nncf required feature.
    # one example is ReduceLROnPlateauLrUpdaterHook
    out = fn(runner, *args, **kwargs)
    setattr(runner, "all_metrics", deepcopy(runner.log_buffer.output))
    return out


NNCF_PATCHER.patch("mmcv.runner.EvalHook.evaluate", _evaluation_wrapper)
NNCF_PATCHER.patch("otx.mpa.modules.hooks.eval_hook.CustomEvalHook.evaluate", _evaluation_wrapper)

NNCF_PATCHER.patch(
    "otx.mpa.modules.hooks.recording_forward_hooks.FeatureVectorHook.func",
    no_nncf_trace_wrapper,
)
NNCF_PATCHER.patch(
    "otx.mpa.modules.hooks.recording_forward_hooks.ActivationMapHook.func",
    no_nncf_trace_wrapper,
)
NNCF_PATCHER.patch(
    "otx.mpa.modules.hooks.recording_forward_hooks.ReciproCAMHook.func",
    no_nncf_trace_wrapper,
)



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from logging import Logger

import pytest
import torch.nn as nn
from mmcv.runner import IterLoader
from torch.utils.data.dataloader import DataLoader

from otx.algorithms.common.adapters.mmcv.runner import (
    EpochRunnerWithCancel,
    IterBasedRunnerWithCancel,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
    load_test_dataset,
)


class TestRunnersInputParamsValidation:
    def iter_based_runner(self):
        return IterBasedRunnerWithCancel(model=self.MockModel(), logger=Logger(name="test logger"))

    @staticmethod
    def data_loader():
        dataset = load_test_dataset()[0]
        return DataLoader(dataset)

    class MockModel(nn.Module):
        @staticmethod
        def train_step():
            pass

    @e2e_pytest_unit
    def test_epoch_runner_with_cancel_train_params_validation(self):
        """
        <b>Description:</b>
        Check EpochRunnerWithCancel object "train" method input parameters validation

        <b>Input data:</b>
        EpochRunnerWithCancel object. "data_loader" non DataLoader object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "train" method
        """
        runner = EpochRunnerWithCancel(model=self.MockModel(), logger=Logger(name="test logger"))
        with pytest.raises(ValueError):
            runner.train(data_loader="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_iter_based_runner_with_cancel_main_loop_params_validation(self):
        """
        <b>Description:</b>
        Check IterBasedRunnerWithCancel object "main_loop" method input parameters validation

        <b>Input data:</b>
        IterBasedRunnerWithCancel object. "main_loop" method unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "main_loop" method
        """
        data_loader = self.data_loader()
        iter_loader = IterLoader(data_loader)
        runner = self.iter_based_runner()
        correct_values_dict = {
            "workflow": [("train", 1)],
            "iter_loaders": [iter_loader],
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "workflow" parameter
            ("workflow", unexpected_int),
            # Unexpected integer is specified as nested workflow
            ("workflow", [("train", 1), unexpected_int]),
            # Unexpected integer is specified as "iter_loaders" parameter
            ("iter_loaders", unexpected_int),
            # Unexpected integer is specified as nested iter_loader
            ("iter_loaders", [iter_loader, unexpected_int]),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=runner.main_loop,
        )

    @e2e_pytest_unit
    def test_iter_based_runner_with_cancel_run_params_validation(self):
        """
        <b>Description:</b>
        Check IterBasedRunnerWithCancel object "run" method input parameters validation

        <b>Input data:</b>
        IterBasedRunnerWithCancel object. "run" method unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "run" method
        """
        data_loader = self.data_loader()
        runner = self.iter_based_runner()
        correct_values_dict = {
            "data_loaders": [data_loader],
            "workflow": [("train", 1)],
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "data_loaders" parameter
            ("data_loaders", unexpected_int),
            # Unexpected integer is specified as nested data_loader
            ("data_loaders", [data_loader, unexpected_int]),
            # Unexpected integer is specified as "workflow" parameter
            ("workflow", unexpected_int),
            # Unexpected integer is specified as nested workflow
            ("workflow", [("train", 1), unexpected_int]),
            # Unexpected string is specified as "max_iters" parameter
            ("max_iters", "unexpected string"),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=runner.run,
        )



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
import os.path as osp
import tempfile

import pytest
import torch

from otx.algorithms.classification.tasks import (  # ClassificationOpenVINOTask,
    ClassificationInferenceTask,
    ClassificationTrainTask,
)
from otx.algorithms.detection.tasks import (
    DetectionInferenceTask,
    DetectionTrainTask,
    OpenVINODetectionTask,
)
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.model import ModelEntity
from otx.api.entities.result_media import ResultMediaEntity
from otx.api.entities.train_parameters import TrainParameters
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.cli.utils.io import read_model, save_model_data
from tests.integration.api.classification.test_api_classification import (
    DEFAULT_CLS_TEMPLATE_DIR,
    ClassificationTaskAPIBase,
)
from tests.integration.api.detection.test_api_detection import (
    DEFAULT_DET_TEMPLATE_DIR,
    DetectionTaskAPIBase,
)
from tests.test_suite.e2e_test_system import e2e_pytest_api

torch.manual_seed(0)

assert_text_explain_all = "The number of saliency maps should be equal to the number of all classes."
assert_text_explain_predicted = "The number of saliency maps should be equal to the number of predicted classes."


def saliency_maps_check(
    predicted_dataset, task_labels, raw_sal_map_shape, processed_saliency_maps=False, only_predicted=True
):
    for data_point in predicted_dataset:
        saliency_map_counter = 0
        metadata_list = data_point.get_metadata()
        for metadata in metadata_list:
            if isinstance(metadata.data, ResultMediaEntity):
                if metadata.data.type == "saliency_map":
                    saliency_map_counter += 1
                    if processed_saliency_maps:
                        assert metadata.data.numpy.ndim == 3
                        assert metadata.data.numpy.shape == (data_point.height, data_point.width, 3)
                    else:
                        assert metadata.data.numpy.ndim == 2
                        assert metadata.data.numpy.shape == raw_sal_map_shape
        if only_predicted:
            assert saliency_map_counter == len(data_point.annotation_scene.get_labels()), assert_text_explain_predicted
        else:
            assert saliency_map_counter == len(task_labels), assert_text_explain_all


class TestOVClsXAIAPI(ClassificationTaskAPIBase):
    ref_raw_saliency_shapes = {
        "EfficientNet-B0": (7, 7),
    }

    @e2e_pytest_api
    @pytest.mark.parametrize(
        "multilabel,hierarchical",
        [(False, False), (True, False), (False, True)],
        ids=["multiclass", "multilabel", "hierarchical"],
    )
    def test_inference_xai(self, multilabel, hierarchical):
        with tempfile.TemporaryDirectory() as temp_dir:
            hyper_parameters, model_template = self.setup_configurable_parameters(DEFAULT_CLS_TEMPLATE_DIR, num_iters=1)
            task_environment, dataset = self.init_environment(
                hyper_parameters, model_template, multilabel, hierarchical, 20
            )

            # Train and save a model
            task = ClassificationTrainTask(task_environment=task_environment)
            train_parameters = TrainParameters()
            output_model = ModelEntity(
                dataset,
                task_environment.get_model_configuration(),
            )
            task.train(dataset, output_model, train_parameters)
            save_model_data(output_model, temp_dir)

            for processed_saliency_maps, only_predicted in [[True, False], [False, True]]:
                task_environment, dataset = self.init_environment(
                    hyper_parameters, model_template, multilabel, hierarchical, 20
                )

                # Infer torch model
                task = ClassificationInferenceTask(task_environment=task_environment)
                inference_parameters = InferenceParameters(
                    is_evaluation=False,
                    process_saliency_maps=processed_saliency_maps,
                    explain_predicted_classes=only_predicted,
                )
                predicted_dataset = task.infer(dataset.with_empty_annotations(), inference_parameters)

                # Check saliency maps torch task
                task_labels = output_model.configuration.get_label_schema().get_labels(include_empty=False)
                saliency_maps_check(
                    predicted_dataset,
                    task_labels,
                    self.ref_raw_saliency_shapes[model_template.name],
                    processed_saliency_maps=processed_saliency_maps,
                    only_predicted=only_predicted,
                )

            # # TODO(negvet): get it back (IR infer does not work)
            # # Save OV IR model
            # task._model_ckpt = osp.join(temp_dir, "weights.pth")
            # exported_model = ModelEntity(None, task_environment.get_model_configuration())
            # task.export(ExportType.OPENVINO, exported_model)
            # os.makedirs(temp_dir, exist_ok=True)
            # save_model_data(exported_model, temp_dir)
            #
            # # Infer OV IR model
            # load_weights_ov = osp.join(temp_dir, "openvino.xml")
            # task_environment.model = read_model(task_environment.get_model_configuration(), load_weights_ov, None)
            # task = ClassificationOpenVINOTask(task_environment=task_environment)
            # predicted_dataset_ov = task.infer(
            #     dataset.with_empty_annotations(),
            #     InferenceParameters(is_evaluation=False),
            # )
            #
            # # Check saliency maps OV task
            # saliency_maps_check(predicted_dataset_ov, task_labels, self.ref_raw_saliency_shapes[model_template.name])


class TestOVDetXAIAPI(DetectionTaskAPIBase):
    ref_raw_saliency_shapes = {
        "ATSS": (4, 4),
        "SSD": (13, 13),
        "YOLOX": (13, 13),
    }

    @e2e_pytest_api
    def test_inference_xai(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            hyper_parameters, model_template = self.setup_configurable_parameters(DEFAULT_DET_TEMPLATE_DIR, num_iters=2)
            detection_environment, dataset = self.init_environment(hyper_parameters, model_template, 10)

            train_task = DetectionTrainTask(task_environment=detection_environment)
            trained_model = ModelEntity(
                dataset,
                detection_environment.get_model_configuration(),
            )
            train_task.train(dataset, trained_model, TrainParameters())
            save_model_data(trained_model, temp_dir)

            # Infer torch model
            detection_environment.model = trained_model
            inference_task = DetectionInferenceTask(task_environment=detection_environment)
            predicted_dataset = inference_task.infer(dataset.with_empty_annotations())

            # Check saliency maps torch task
            task_labels = trained_model.configuration.get_label_schema().get_labels(include_empty=False)
            saliency_maps_check(predicted_dataset, task_labels, self.ref_raw_saliency_shapes[model_template.name])

            # Save OV IR model
            inference_task._model_ckpt = osp.join(temp_dir, "weights.pth")
            exported_model = ModelEntity(None, detection_environment.get_model_configuration())
            inference_task.export(ExportType.OPENVINO, exported_model)
            os.makedirs(temp_dir, exist_ok=True)
            save_model_data(exported_model, temp_dir)

            # Infer OV IR model
            load_weights_ov = osp.join(temp_dir, "openvino.xml")
            detection_environment.model = read_model(
                detection_environment.get_model_configuration(), load_weights_ov, None
            )
            task = OpenVINODetectionTask(task_environment=detection_environment)
            predicted_dataset_ov = task.infer(
                dataset.with_empty_annotations(),
                InferenceParameters(is_evaluation=False),
            )

            # Check saliency maps OV task
            saliency_maps_check(predicted_dataset_ov, task_labels, self.ref_raw_saliency_shapes[model_template.name])



"""This module contains the mapper for datetime."""

#
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import datetime
from typing import Union

from otx.api.utils.time_utils import now


class DatetimeMapper:
    """This class maps a `datetime.datetime` entity to a string, and vice versa."""

    @staticmethod
    def forward(instance: datetime.datetime) -> str:
        """Serializes datetime to str."""

        return instance.strftime("%Y-%m-%dT%H:%M:%S.%f")

    @staticmethod
    def backward(instance: Union[None, str]) -> datetime.datetime:
        """Deserializes datetime from str or create new one if it is None."""

        if isinstance(instance, str):
            modification_date = datetime.datetime.strptime(instance, "%Y-%m-%dT%H:%M:%S.%f")
            return modification_date.replace(tzinfo=datetime.timezone.utc)

        return now()



"""NNCF Task for OTX Classification."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from functools import partial
from typing import List, Optional

import otx.algorithms.classification.adapters.mmcls.nncf.patches  # noqa: F401  # pylint: disable=unused-import
import otx.algorithms.classification.adapters.mmcls.nncf.registers  # noqa: F401  # pylint: disable=unused-import
from otx.algorithms.classification.adapters.mmcls.nncf.builder import (
    build_nncf_classifier,
)
from otx.algorithms.common.tasks.nncf_base import NNCFBaseTask
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.metrics import (
    CurveMetric,
    LineChartInfo,
    LineMetricsGroup,
    MetricsGroup,
    Performance,
    ScoreMetric,
)
from otx.api.entities.model import ModelEntity  # ModelStatus
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.mpa.utils.logger import get_logger

from .inference import ClassificationInferenceTask

logger = get_logger()


class ClassificationNNCFTask(NNCFBaseTask, ClassificationInferenceTask):  # pylint: disable=too-many-ancestors
    """ClassificationNNCFTask."""

    def _initialize_post_hook(self, options=None):
        super()._initialize_post_hook(options)

        export = options.get("export", False)
        options["model_builder"] = partial(
            self.model_builder,
            nncf_model_builder=build_nncf_classifier,
            return_compression_ctrl=False,
            is_export=export,
        )

    def _optimize(
        self,
        dataset: DatasetEntity,
        optimization_parameters: Optional[OptimizationParameters] = None,
    ):
        results = self._run_task(
            "ClsTrainer",
            mode="train",
            dataset=dataset,
            parameters=optimization_parameters,
        )
        return results

    def _optimize_post_hook(
        self,
        dataset: DatasetEntity,
        output_model: ModelEntity,
    ):
        # Get training metrics group from learning curves
        training_metrics, final_acc = self._generate_training_metrics_group(self._learning_curves)
        performance = Performance(
            score=ScoreMetric(value=final_acc, name="accuracy"),
            dashboard_metrics=training_metrics,
        )

        logger.info(f"Final model performance: {str(performance)}")
        output_model.performance = performance

    def _generate_training_metrics_group(self, learning_curves):
        """Parses the classification logs to get metrics from the latest training run.

        :return output List[MetricsGroup]
        """
        output: List[MetricsGroup] = []

        if self._multilabel:
            metric_key = "val/accuracy-mlc"
        elif self._hierarchical:
            metric_key = "val/MHAcc"
        else:
            metric_key = "val/accuracy_top-1"

        # Learning curves
        best_acc = -1
        if learning_curves is None:
            return output

        for key, curve in learning_curves.items():
            metric_curve = CurveMetric(xs=curve.x, ys=curve.y, name=key)
            if key == metric_key:
                best_acc = max(curve.y)
            visualization_info = LineChartInfo(name=key, x_axis_label="Timestamp", y_axis_label=key)
            output.append(LineMetricsGroup(metrics=[metric_curve], visualization_info=visualization_info))

        return output, best_acc



"""Utilities and use cases built on top of OTX API.

.. automodule:: otx.api.usecases.adapters
   :members:
   :undoc-members:

"""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0



"""Unit Test for otx.algorithms.detection.adapters.mmdet.data.dataset."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import numpy as np
import pytest

from otx.algorithms.detection.adapters.mmdet.data.dataset import OTXDetDataset
from otx.api.entities.label import Domain
from otx.api.entities.model_template import TaskType
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.detection.test_helpers import (
    MockPipeline,
    generate_det_dataset,
)


class TestOTXDetDataset:
    """
    Test OTXDetDataset class.
    1. Test _DataInfoProxy
    2. Test prepare_train_img
    3. Test prepare_test_img
    4. Test get_ann_info
    5. Test evaluate
    """

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.dataset = dict()
        for task_type in [TaskType.DETECTION, TaskType.INSTANCE_SEGMENTATION]:
            self.dataset[task_type] = generate_det_dataset(task_type=task_type)
        self.pipeline = []

    @e2e_pytest_unit
    @pytest.mark.parametrize("task_type", [TaskType.DETECTION, TaskType.INSTANCE_SEGMENTATION])
    def test_DataInfoProxy(self, task_type):
        """Test _DataInfoProxy Class."""
        otx_dataset, labels = self.dataset[task_type]
        proxy = OTXDetDataset._DataInfoProxy(otx_dataset, labels)
        sample = proxy[0]
        assert "dataset_item" in sample
        assert "width" in sample
        assert "height" in sample
        assert "index" in sample
        assert "ann_info" in sample
        assert "ignored_labels" in sample

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "task_type, domain",
        [(TaskType.DETECTION, Domain.DETECTION), (TaskType.INSTANCE_SEGMENTATION, Domain.INSTANCE_SEGMENTATION)],
    )
    def test_prepare_train_img(self, task_type, domain) -> None:
        """Test prepare_train_img method"""
        otx_dataset, labels = self.dataset[task_type]
        dataset = OTXDetDataset(otx_dataset, labels, self.pipeline, domain, test_mode=False)
        img = dataset.prepare_train_img(0)
        assert isinstance(img, dict)
        assert "dataset_item" in img
        assert "bbox_fields" in img
        assert "mask_fields" in img
        assert "seg_fields" in img

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "task_type, domain",
        [(TaskType.DETECTION, Domain.DETECTION), (TaskType.INSTANCE_SEGMENTATION, Domain.INSTANCE_SEGMENTATION)],
    )
    def test_prepare_test_img(self, task_type, domain) -> None:
        """Test prepare_test_img method"""
        otx_dataset, labels = self.dataset[task_type]
        dataset = OTXDetDataset(otx_dataset, labels, self.pipeline, domain, test_mode=True)
        img = dataset.prepare_test_img(0)
        assert isinstance(img, dict)
        assert "dataset_item" in img
        assert "bbox_fields" in img
        assert "mask_fields" in img
        assert "seg_fields" in img

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "task_type, domain",
        [(TaskType.DETECTION, Domain.DETECTION), (TaskType.INSTANCE_SEGMENTATION, Domain.INSTANCE_SEGMENTATION)],
    )
    def test_get_ann_info(self, task_type, domain) -> None:
        """Test get_ann_info method"""
        otx_dataset, labels = self.dataset[task_type]
        dataset = OTXDetDataset(otx_dataset, labels, self.pipeline, domain)
        dataset.pipeline = MockPipeline()
        ann_info = dataset.get_ann_info(0)
        assert isinstance(ann_info, dict)
        assert "bboxes" in ann_info
        assert "masks" in ann_info
        assert "labels" in ann_info

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "task_type, domain",
        [(TaskType.DETECTION, Domain.DETECTION), (TaskType.INSTANCE_SEGMENTATION, Domain.INSTANCE_SEGMENTATION)],
    )
    @pytest.mark.parametrize("metric", ["mAP"])
    @pytest.mark.parametrize("logger", ["silent", None])
    def test_evaluate(self, task_type, domain, metric, logger) -> None:
        """Test evaluate method for detection and instance segmentation"""
        otx_dataset, labels = self.dataset[task_type]
        dataset = OTXDetDataset(otx_dataset, labels, self.pipeline, domain)
        dataset.pipeline = MockPipeline()
        sample = dataset[0]
        if task_type == TaskType.DETECTION:
            results = [[np.random.rand(1, 5)]]
        elif task_type == TaskType.INSTANCE_SEGMENTATION:
            results = [
                (
                    [np.random.rand(1, 5)],
                    [[{"size": [sample["width"], sample["height"]], "counts": "some counts"}]],
                )
            ]
        eval_results = dataset.evaluate(results, metric, logger)
        assert isinstance(eval_results, dict)
        assert metric in eval_results



"""Unit Test for otx.algorithms.action.tools.sample_classification."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv.utils import Config

from otx.algorithms.action.tools.sample_classification import (
    load_test_dataset,
    main,
    parse_args,
)
from otx.algorithms.common.configs.training_base import TrainType
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import Domain
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model_template import TaskType
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.action.test_helpers import (
    generate_action_cls_otx_dataset,
    generate_labels,
)


@e2e_pytest_unit
def test_parse_args(mocker) -> None:
    """Test parse_args function."""

    class MockArgParser:
        def __init__(self, description):
            self.description = description

        def add_argument(self, name, *args, **kwargs):
            setattr(self, name.split("--")[-1], True)

        def parse_args(self):
            return self

    mocker.patch("otx.algorithms.action.tools.sample_classification.argparse.ArgumentParser", side_effect=MockArgParser)
    parser = parse_args()
    assert parser.template_file_path is not None
    assert parser.export is not None


@e2e_pytest_unit
def test_load_test_dataset() -> None:
    """Test laod_test_dataset function."""

    class MockTemplate:
        task_type = TaskType.ACTION_CLASSIFICATION
        train_type = TrainType.INCREMENTAL.value

    dataset, label_schema = load_test_dataset(MockTemplate())
    isinstance(dataset, DatasetEntity)
    isinstance(label_schema, LabelSchemaEntity)


@e2e_pytest_unit
def test_main(mocker) -> None:
    """Test main function."""

    class MockArgs:
        template_file_path = "dummy_path"
        export = False

    class MockTaskEnvironment:
        def __init__(self, *args, **kwargs):
            self.model = None

        def get_model_configuration(self):
            return True

    class MockTestCls:
        def __init__(self, task_environment):
            pass

        def train(self, dataset, output_model):
            pass

        def infer(self, dataset, params):
            return dataset

        def evaluate(self, resultset):
            resultset.performance = 1.0

        def export(self, export_type, model):
            return model

        def optimize(self, optimization_type, dataset, modle, params):
            pass

    mocker.patch(
        "otx.algorithms.action.tools.sample_classification.parse_model_template",
        return_value=Config(
            {
                "hyper_parameters": {"data": "dummy_data"},
                "entrypoints": {"base": "dummy_base", "openvino": "dummy_base"},
            }
        ),
    )
    mocker.patch(
        "otx.algorithms.action.tools.sample_classification.load_test_dataset",
        return_value=(
            generate_action_cls_otx_dataset(3, 3, generate_labels(3, Domain.ACTION_CLASSIFICATION)),
            "dummy_label_schema",
        ),
    )
    mocker.patch(
        "otx.algorithms.action.tools.sample_classification.create",
        return_value=Config({"learning_parameters": {"num_iters": 4}}),
    )
    mocker.patch("otx.algorithms.action.tools.sample_classification.TaskEnvironment", side_effect=MockTaskEnvironment)
    mocker.patch("otx.algorithms.action.tools.sample_classification.get_task_class", return_value=MockTestCls)
    main(MockArgs())

    MockArgs.export = True
    main(MockArgs())



"""Runner with cancel for common OTX algorithms."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# Is based on
# * https://github.com/open-mmlab/mmcv/blob/master/mmcv/runner/epoch_based_runner.py
# * https://github.com/open-mmlab/mmcv/blob/master/mmcv/runner/iter_based_runner.py

import time
import warnings
from typing import List, Optional, Sequence

import mmcv
import torch.distributed as dist
from mmcv.runner import (
    RUNNERS,
    EpochBasedRunner,
    IterBasedRunner,
    IterLoader,
    get_dist_info,
)
from mmcv.runner.utils import get_host_info
from torch.utils.data.dataloader import DataLoader

from otx.api.utils.argument_checks import check_input_parameters_type


# pylint: disable=too-many-instance-attributes, attribute-defined-outside-init
@RUNNERS.register_module()
class EpochRunnerWithCancel(EpochBasedRunner):
    """Simple modification to EpochBasedRunner to allow cancelling the training during an epoch.

    A stopping hook should set the runner.should_stop flag to True if stopping is required.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.should_stop = False
        _, world_size = get_dist_info()
        self.distributed = world_size > 1
        self.save_ema_model = False

    def stop(self) -> bool:
        """Returning a boolean to break the training loop.

        This method supports distributed training by broadcasting should_stop to other ranks
        :return: a cancellation bool
        """
        broadcast_obj = [False]
        if self.rank == 0 and self.should_stop:
            broadcast_obj = [True]

        if self.distributed:
            dist.broadcast_object_list(broadcast_obj, src=0)
        if broadcast_obj[0]:
            self._max_epochs = self.epoch
        return broadcast_obj[0]

    @check_input_parameters_type()
    def train(self, data_loader: DataLoader, **kwargs):
        """Train call hook."""
        self.model.train()
        self.mode = "train"
        self.data_loader = data_loader
        self._max_iters = self._max_epochs * len(self.data_loader)
        self.call_hook("before_train_epoch")
        if self.distributed:
            time.sleep(2)  # Prevent possible multi-gpu deadlock during epoch transition
        for i, data_batch in enumerate(self.data_loader):
            self._inner_iter = i
            self.call_hook("before_train_iter")
            self.run_iter(data_batch, train_mode=True, **kwargs)
            self.call_hook("after_train_iter")
            if self.stop():
                break
            self._iter += 1
            self.save_ema_model = False  # revert ema status before new iter
        self.call_hook("after_train_epoch")
        self.stop()
        self._epoch += 1


@RUNNERS.register_module()
class IterBasedRunnerWithCancel(IterBasedRunner):
    """Runner With Cancel for early-stopping (Iter based).

    Simple modification to IterBasedRunner to allow cancelling the training. The cancel training hook
    should set the runner.should_stop flag to True if stopping is required.

    # TODO: Implement cancelling of training via keyboard interrupt signal, instead of should_stop
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.should_stop = False

    @check_input_parameters_type()
    def main_loop(self, workflow: List[tuple], iter_loaders: Sequence[IterLoader], **kwargs):
        """Main loop function in IterBasedRunnerWithCancel."""
        while self.iter < self._max_iters:
            for i, flow in enumerate(workflow):
                self._inner_iter = 0
                mode, iters = flow
                if not isinstance(mode, str) or not hasattr(self, mode):
                    raise ValueError(f'runner has no method named "{mode}" to run a workflow')
                iter_runner = getattr(self, mode)
                for _ in range(iters):
                    if mode == "train" and self.iter >= self._max_iters:
                        break
                    iter_runner(iter_loaders[i], **kwargs)
                    if self.should_stop:
                        return

    @check_input_parameters_type()
    def run(self, data_loaders: Sequence[DataLoader], workflow: List[tuple], max_iters: Optional[int] = None, **kwargs):
        """Function of main run."""
        assert isinstance(data_loaders, list)
        assert mmcv.is_list_of(workflow, tuple)
        assert len(data_loaders) == len(workflow)
        if max_iters is not None:
            warnings.warn(
                "setting max_iters in run is deprecated, please set max_iters in runner_config",
                DeprecationWarning,
            )
            self._max_iters = max_iters
        assert self._max_iters is not None, "max_iters must be specified during instantiation"

        work_dir = self.work_dir if self.work_dir is not None else "NONE"
        self.logger.info("Start running, host: %s, work_dir: %s", get_host_info(), work_dir)
        self.logger.info("workflow: %s, max: %d iters", workflow, self._max_iters)
        self.call_hook("before_run")

        iter_loaders = [IterLoader(x) for x in data_loaders]

        self.call_hook("before_epoch")

        self.should_stop = False
        self.main_loop(workflow, iter_loaders, **kwargs)
        self.should_stop = False

        # time.sleep(1)  # wait for some hooks like loggers to finish
        self.call_hook("after_epoch")
        self.call_hook("after_run")



"""Adapters of classification - mmcls."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


from .data import OTXClsDataset, SelfSLDataset
from .models import BYOL, ConstrastiveHead, SelfSLMLP

# fmt: off
# isort: off
# FIXME: openvino pot library adds stream handlers to root logger
# which makes annoying duplicated logging
from mmcls.utils import get_root_logger  # pylint: disable=wrong-import-order
get_root_logger().propagate = False
# isort:on
# fmt: on

__all__ = [
    "OTXClsDataset",
    "SelfSLDataset",
    "BYOL",
    "SelfSLMLP",
    "ConstrastiveHead",
]



"""This module implements the ResultMediaEntity."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Optional

import numpy as np

from otx.api.entities.annotation import Annotation, AnnotationSceneEntity
from otx.api.entities.label import LabelEntity
from otx.api.entities.metadata import IMetadata
from otx.api.entities.shapes.rectangle import Rectangle


# pylint: disable=too-many-instance-attributes; Requires refactor
class ResultMediaEntity(IMetadata):
    """Represents a media (e.g. an image which was generated by a task).

    For instance, a `ResultMediaEntity` could be an attention map generated by a classification task.

    The result media contains media data, which is associated with a
    `otx.api.entities.annotation.AnnotationSceneEntity` and related to an optional
    `otx.api.entities.label.LabelEntity`.

    Example:
        >>> from otx.api.entities.annotation import (
            Annotation,
            AnnotationSceneEntity,
            AnnotationSceneKind,
            )
        >>> from otx.api.entities.id import ID
        >>> from otx.api.entities.label import Domain, LabelEntity
        >>> from otx.api.entities.result_media import ResultMediaEntity
        >>> from otx.api.entities.scored_label import LabelSource, ScoredLabel
        >>> from otx.api.entities.shapes.rectangle import Rectangle

        >>> source = LabelSource(
                user_id="user_entity", model_id=ID("efficientnet"), model_storage_id=ID("efficientnet-storage")
                )
        >>> falcon_label = LabelEntity(name="Falcon", domain=Domain.DETECTION)
        >>> eagle_label = LabelEntity(name="Eagle", domain=Domain.DETECTION)
        >>> falcon_bb = Rectangle(x1=0.0, y1=0.0, x2=0.5, y2=0.5)
        >>> falcon_scored_label = ScoredLabel(label=falcon_label, probability=0.9, label_source=source)
        >>> eagle_bb = Rectangle(x1=0.2, y1=0.2, x2=0.8, y2=0.8)
        >>> eagle_scored_label = ScoredLabel(label=eagle_label, probability=0.6, label_source=source)
        >>> annotation_scene = AnnotationSceneEntity(
                 annotations=[
                     Annotation(shape=falcon_bb, labels=[falcon_scored_label]),
                     Annotation(shape=eagle_bb, labels=[eagle_scored_label]),
                 ], kind=AnnotationSceneKind.PREDICTION
             )
        >>> ResultMediaEntity(
                name="Model Predictions",
                type="Bounding Box Annotations",
                annotation_scene=annotation_scene,
                numpy=image_array
            )

    Args:
        name (str): Name.
        type (str): The type of data (e.g. Attention map). This type is descriptive.
        annotation_scene (AnnotationScene Entity): Associated annotation which was generated by the task
                                alongside this media.
        numpy (np.ndarray): The data as a numpy array.
        roi (Optional[Annotation]): The ROI covered by this media. If null, assume the entire image. Defaults to None.
        label (Optional[LabelEntity]): A label associated with this media. Defaults to None.
    """

    # pylint: disable=redefined-builtin, too-many-arguments;
    def __init__(
        self,
        name: str,
        type: str,
        annotation_scene: AnnotationSceneEntity,
        numpy: np.ndarray,
        roi: Optional[Annotation] = None,
        label: Optional[LabelEntity] = None,
    ):
        self.name = name
        self.type = type
        self.annotation_scene = annotation_scene
        self.roi = Annotation(Rectangle.generate_full_box(), labels=[]) if roi is None else roi
        self.label = label
        self.numpy = numpy

    def __repr__(self):
        """Returns a string with all the attributes of the ResultMediaEntity."""
        return (
            "ResultMediaEntity("
            f"name={self.name}, "
            f"type={self.type}, "
            f"annotation_scene={self.annotation_scene}, "
            f"roi={self.roi}, "
            f"label={self.label})"
        )

    @property
    def width(self) -> int:
        """Returns the width of the result media."""
        return self.numpy.shape[1]

    @property
    def height(self) -> int:
        """Returns the height of the result media."""
        return self.numpy.shape[0]

    @property
    def numpy(self) -> np.ndarray:
        """Returns the data."""
        return self._numpy

    @numpy.setter
    def numpy(self, value):
        self._numpy = value

    def __eq__(self, other):
        """Checks if the annotation_scene and roi matches with the other ResultMediaEntity."""
        if isinstance(other, ResultMediaEntity):
            return self.annotation_scene == other.annotation_scene and self.roi == other.roi
        return False



"""Unit Test for otx.algorithms.action.adapters.openvino.model_wrappers.openvino_models."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Dict

import numpy as np
import pytest
from mmcv.utils import Config
from openvino.model_zoo.model_api.adapters import OpenvinoAdapter

from otx.algorithms.action.adapters.openvino.model_wrappers.openvino_models import (
    OTXOVActionCls,
    OTXOVActionDet,
    get_multiclass_predictions,
    softmax_numpy,
)
from otx.api.entities.label import Domain
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.action.test_helpers import (
    generate_action_cls_otx_dataset,
    generate_labels,
)


class MockOpenvinoAdapter(OpenvinoAdapter):
    """Mock class for OpenvinoAdapter."""

    def __init__(self):
        pass


class MockOTXOVActionCls(OTXOVActionCls):
    """Mock class for OTXOVActionCls."""

    def __init__(self, *args):
        self.inputs: Dict[str, np.ndarray] = {
            "cls_data": np.ndarray([1, 1, 3, 8, 256, 256]),
            "det_data": np.ndarray([1, 3, 8, 256, 256]),
            "cls_info": np.ndarray([1, 1, 1, 1]),
            "dummy": np.ndarray([1, 1, 1]),
        }

        self.outputs: Dict[str, Config] = {
            "logits": Config({"names": "cls_layer"}),
            "gt_bboxes": Config({"names": "reg_layer"}),
            "gt_labels": Config({"names": "cls_layer"}),
        }
        super().__init__(MockOpenvinoAdapter)


class MockOTXOVActionDet(OTXOVActionDet):
    """Mock class for OTXOVActionDet."""

    def __init__(self, *args):
        self.inputs: Dict[str, np.ndarray] = {
            "cls_data": np.ndarray([1, 1, 3, 8, 256, 256]),
            "det_data": np.ndarray([1, 3, 8, 256, 256]),
            "cls_info": np.ndarray([1, 1, 1, 1]),
            "dummy": np.ndarray([1, 1, 1]),
        }

        self.outputs: Dict[str, Config] = {
            "logits": Config({"names": "cls_layer"}),
            "gt_bboxes": Config({"names": "reg_layer"}),
            "gt_labels": Config({"names": "cls_layer"}),
        }
        super().__init__(MockOpenvinoAdapter)


@e2e_pytest_unit
def test_softmax_numpy() -> None:
    """Test softmax_numpy function.

    It checks argmax of inputs and outputs
    """

    x = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
    out = softmax_numpy(x)
    assert np.argmax(x) == np.argmax(out)


@e2e_pytest_unit
def test_get_multiclass_predictions(mocker) -> None:
    """Test get_multiclass_predictions function.

    It checks argmax and max of inputs and outputs
    """

    inputs = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
    outputs = get_multiclass_predictions(inputs, False)
    assert 4 == outputs[0][0]
    assert 5.0 == outputs[0][1]

    mocker.patch(
        "otx.algorithms.action.adapters.openvino.model_wrappers.openvino_models.softmax_numpy", return_value=inputs
    )
    outputs = get_multiclass_predictions(inputs, False)
    assert 4 == outputs[0][0]
    assert 5.0 == outputs[0][1]


class TestOTXOVActionCls:
    """Test OTXOVActionCls class.

    1. Test __init__ function
    <Steps>
        1. Check model's input, output name
        2. Check model's input's dimension
    2. Test preprocess function
    <Steps>
        1. Generate sample items: List[DatasetItemEntity]
        2. Check pre-processed inputs
            1. Check inputs' dimension
            2. Check meta information
    3. Test postprocess function
    <Steps>
        1. Generate sample output
        2. Check postprocess function's output return's argmax
    """

    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        mocker.patch(
            "otx.algorithms.action.adapters.openvino.model_wrappers.openvino_models.Model.__init__",
            return_value=True,
        )
        self.model = MockOTXOVActionCls()

    @e2e_pytest_unit
    def test_init(self, mocker) -> None:
        """Test __init__ function."""
        assert self.model.image_blob_names == ["cls_data"]
        assert self.model.image_info_blob_names == ["cls_info"]
        assert self.model.out_layer_name == "logits"
        assert self.model.n == 1
        assert self.model.c == 3
        assert self.model.t == 8
        assert self.model.h == 256
        assert self.model.w == 256

    @e2e_pytest_unit
    def test_preprocess(self) -> None:
        """Test preprocess function."""

        labels = generate_labels(1, Domain.ACTION_CLASSIFICATION)
        items = generate_action_cls_otx_dataset(1, 10, labels)._items

        dict_inputs, meta = self.model.preprocess(items)
        assert dict_inputs["cls_data"].shape == (1, 1, 3, 10, 256, 256)
        assert meta["original_shape"] == (256, 256, 3)
        assert meta["resized_shape"] == (1, 3, 10, 256, 256)

    @e2e_pytest_unit
    def test_postprocess(self) -> None:
        """Test postprocess function."""

        sample_output = {"logits": np.array([1.0, 2.0, 3.0, 4.0, 5.0])}
        out = self.model.postprocess(sample_output, meta={"Any": "Any"})
        assert out[0][0] == 4


class TestOTXOVActionDet:
    """Test OTXOVActionDet class.

    1. Test __init__ function
    <Steps>
        1. Check model's input, output name
        2. Check model's input's dimension
    2. Test preprocess function
    <Steps>
        1. Generate sample items: List[DatasetItemEntity]
        2. Check pre-processed inputs
            1. Check inputs' dimension
            2. Check meta information
    3. Test postprocess function
    <Steps>
        1. Generate sample output
        2. Check postprocess function output's id, score, bbox
    """

    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        mocker.patch(
            "otx.algorithms.action.adapters.openvino.model_wrappers.openvino_models.Model.__init__",
            return_value=True,
        )
        self.model = MockOTXOVActionDet()

    @e2e_pytest_unit
    def test_init(self, mocker) -> None:
        """Test __init__ function."""
        assert self.model.image_blob_names == ["det_data"]
        assert self.model.out_layer_names == {"bboxes": "gt_bboxes", "labels": "gt_labels"}
        assert self.model.n == 1
        assert self.model.c == 3
        assert self.model.t == 8
        assert self.model.h == 256
        assert self.model.w == 256

    @e2e_pytest_unit
    def test_preprocess(self) -> None:
        """Test preprocess function."""

        labels = generate_labels(1, Domain.ACTION_CLASSIFICATION)
        items = generate_action_cls_otx_dataset(1, 10, labels)._items

        dict_inputs, meta = self.model.preprocess(items)
        assert dict_inputs["det_data"].shape == (1, 3, 10, 256, 256)
        assert meta["original_shape"] == (256, 256, 3)
        assert meta["resized_shape"] == (1, 3, 10, 256, 256)

    @e2e_pytest_unit
    def test_postprocess(self) -> None:
        """Test postprocess function."""

        sample_output = {"gt_bboxes": np.array([[0, 0, 1, 1]]), "gt_labels": np.array([[0.3, 0.2, 0.1, 0.7]])}
        out = self.model.postprocess(sample_output, meta={"original_shape": (256, 256, 3)})
        # argmax index is 2 because first index is for background
        assert out[0].id == 2
        assert out[0].score == 0.7
        assert out[0].get_coords() == (0, 0, 256, 256)



"""This module implements the Tensor entity."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Tuple

import numpy as np

from otx.api.entities.metadata import IMetadata


class TensorEntity(IMetadata):
    """Represents a metadata of tensor type in OTX.

    Args:
        name: name of metadata
        numpy: the numpy data of the tensor
    """

    def __init__(self, name: str, numpy: np.ndarray):
        self.name = name
        # Copying Numpy array as it points to the same memory address
        self._numpy = np.copy(numpy)

    @property
    def numpy(self) -> np.ndarray:
        """Returns the numpy representation of the tensor."""
        return self._numpy

    @numpy.setter
    def numpy(self, value):
        self._numpy = value

    @property
    def shape(self) -> Tuple[int, ...]:
        """Returns the shape of the tensor."""
        return self._numpy.shape

    def __eq__(self, other):
        """Returns True if the tensors are equal."""
        if isinstance(other, TensorEntity):
            return np.array_equal(self.numpy, other.numpy)
        return False

    def __str__(self):
        """Returns the string representation of the tensor."""
        return f"{self.__class__.__name__}(name={self.name}, shape={self.shape})"

    def __repr__(self):
        """Returns the representation of the tensor."""
        return f"{self.__class__.__name__}(name={self.name})"



"""Model configuration of X3D model for Action Classification Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

num_classes = 400
num_samples = 12
model = dict(
    type="Recognizer3D",
    backbone=dict(type="X3D", gamma_w=1, gamma_b=2.25, gamma_d=2.2),
    cls_head=dict(
        type="X3DHead", in_channels=432, num_classes=num_classes, spatial_type="avg", dropout_ratio=0.5, fc1_bias=False
    ),
    # model training and testing settings
    train_cfg=None,
    test_cfg=dict(average_clips="prob"),
)

evaluation = dict(interval=1, metrics=["top_k_accuracy", "mean_class_accuracy"], final_metric="mean_class_accuracy")

optimizer = dict(
    type="AdamW",
    lr=0.001,
    weight_decay=0.0001,
)

optimizer_config = dict(grad_clip=dict(max_norm=40.0, norm_type=2))
lr_config = dict(policy="step", step=5)
total_epochs = 5

# runtime settings
checkpoint_config = dict(interval=1)
log_config = dict(
    interval=10,
    hooks=[
        dict(type="TextLoggerHook", ignore_last=False),
    ],
)
# runtime settings
log_level = "INFO"
workflow = [("train", 1)]

find_unused_parameters = False
gpu_ids = range(0, 1)

dist_params = dict(backend="nccl")
resume_from = None
load_from = (
    "https://download.openmmlab.com/mmaction/recognition/x3d/facebook/"
    "x3d_m_facebook_16x5x1_kinetics400_rgb_20201027-3f42382a.pth"
)



"""Collections of Dataset utils for common OTX algorithms."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

import glob
import logging
import os
from typing import Any, Dict, Optional, Union

import cv2
import numpy as np

from otx.api.entities.annotation import NullAnnotationSceneEntity
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.image import Image
from otx.api.entities.subset import Subset
from otx.api.utils.argument_checks import (
    IMAGE_FILE_EXTENSIONS,
    DirectoryPathCheck,
    check_input_parameters_type,
)

logger = logging.getLogger(__name__)


@check_input_parameters_type({"file_list_path": DirectoryPathCheck})
def get_unlabeled_filename(base_root: str, file_list_path: str):
    """This method checks and gets image file paths, which are listed in file_list_path.

    The content of file_list_path is expected to specify relative paths of each image file to base_root line by line.
    It returns the list of image filenames only which will compose unlabeled dataset.

    Args:
        base_root (str): path of base root dir where unlabeled images are.
        file_list_path (str) : path of file which contains relative paths of unlabeled data to base_root.

    Returns:
        List[str]: a list of existing image file paths which will be unlabeled data items.
    """

    def is_valid(file_path):
        return file_path.lower().endswith(tuple(IMAGE_FILE_EXTENSIONS))

    with open(file_list_path, "r", encoding="UTF-8") as f:
        file_names = f.read().splitlines()
    unlabeled_files = []
    for fn in file_names:
        file_path = os.path.join(base_root, fn.strip())
        if is_valid(file_path) and os.path.isfile(file_path):
            unlabeled_files.append(file_path)
    return unlabeled_files


@check_input_parameters_type({"data_root_dir": DirectoryPathCheck})
def load_unlabeled_dataset_items(
    data_root_dir: str,
    file_list_path: Optional[str] = None,
):
    """This method loads unlabeled dataset items from images in data_root_dir.

    Args:
        data_root_dir (str): path of base root directory where unlabeled images are.
        file_list_path (str) : path of a file which contains relative paths of unlabeled data to base_root.
        subset (Subset) : Entity subset category
    Returns:
        List[DatasetItemEntity]: a list of unlabeled dataset item entity.
    """
    if file_list_path is not None:
        data_list = get_unlabeled_filename(data_root_dir, file_list_path)

    else:
        data_list = []

        for ext in IMAGE_FILE_EXTENSIONS:
            data_list.extend(glob.glob(f"{data_root_dir}/**/*{ext}", recursive=True))

    dataset_items = []

    for filename in data_list:
        dataset_item = DatasetItemEntity(
            media=Image(file_path=filename),
            annotation_scene=NullAnnotationSceneEntity(),
            subset=Subset.UNLABELED,
        )
        dataset_items.append(dataset_item)
    return dataset_items


def get_dataset(dataset: DatasetEntity, subset: Subset):
    """Get dataset from datasetentity."""
    data = dataset.get_subset(subset)
    return data if len(data) > 0 else None


def get_cls_img_indices(labels, dataset):
    """Function for getting image indices per class.

    Args:
        labels (List[LabelEntity]): List of labels
        dataset(DatasetEntity): dataset entity
    """
    img_indices = {label.name: [] for label in labels}
    for i, item in enumerate(dataset):
        item_labels = item.annotation_scene.get_labels()
        for i_l in item_labels:
            if i_l in labels:
                img_indices[i_l.name].append(i)

    return img_indices


def get_old_new_img_indices(labels, new_classes, dataset):
    """Function for getting old & new indices of dataset.

    Args:
        labels (List[LabelEntity]): List of labels
        new_classes(List[str]): List of new classes
        dataset(DatasetEntity): dataset entity
    """
    ids_old, ids_new = [], []
    _dataset_label_schema_map = {label.name: label for label in labels}
    new_classes = [_dataset_label_schema_map[new_class] for new_class in new_classes]
    for i, item in enumerate(dataset):
        if item.annotation_scene.contains_any(new_classes):
            ids_new.append(i)
        else:
            ids_old.append(i)
    return {"old": ids_old, "new": ids_new}


def get_image(results: Dict[str, Any], cache_dir: str, to_float32=False) -> np.ndarray:
    """Load an image and cache it if it's a training video frame.

    Args:
        results (Dict[str, Any]): A dictionary that contains information about the dataset item.
        cache_dir (str): A directory path where the cached images will be stored.
        to_float32 (bool, optional): A flag indicating whether to convert the image to float32. Defaults to False.

    Returns:
        np.ndarray: The loaded image.
    """

    def is_training_video_frame(subset, media) -> bool:
        return subset.name in ["TRAINING", "VALIDATION"] and "VideoFrame" in repr(media)

    def load_image_from_cache(filename: str, to_float32=False) -> Union[np.ndarray, None]:
        try:
            cached_img = cv2.imread(filename)
            if to_float32:
                cached_img = cached_img.astype(np.float32)
            return cached_img
        except Exception as e:  # pylint: disable=broad-except
            logger.warning(f"Skip loading cached {filename} \nError msg: {e}")
            return None

    def save_image_to_cache(img: np.array, filename: str):
        tmp_filename = filename.replace(".png", "-tmp.png")
        if os.path.exists(filename) or os.path.exists(tmp_filename):  # if image is cached or caching
            return
        try:
            cv2.imwrite(tmp_filename, img=img)
        except Exception as e:  # pylint: disable=broad-except
            logger.warning(f"Skip caching for {filename} \nError msg: {e}")
            return

        if os.path.exists(tmp_filename) and not os.path.exists(filename):
            try:
                os.replace(tmp_filename, filename)
            except Exception as e:  # pylint: disable=broad-except
                os.remove(tmp_filename)
                logger.warning(f"Failed to rename {tmp_filename} -> {filename} \nError msg: {e}")

    subset = results["dataset_item"].subset
    media = results["dataset_item"].media
    if is_training_video_frame(subset, media):
        index = results["index"]
        filename = os.path.join(cache_dir, f"{subset}-{index:06d}.png")
        if os.path.exists(filename):
            loaded_img = load_image_from_cache(filename, to_float32=to_float32)
            if loaded_img is not None:
                return loaded_img

    img = results["dataset_item"].numpy  # this takes long for VideoFrame
    if to_float32:
        img = img.astype(np.float32)

    if is_training_video_frame(subset, media):
        save_image_to_cache(img, filename)

    return img



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
from mmcls.models.builder import CLASSIFIERS

from otx.mpa.utils.logger import get_logger

from .sam_classifier import SAMImageClassifier

logger = get_logger()


@CLASSIFIERS.register_module()
class SemiSLClassifier(SAMImageClassifier):
    """Semi-SL Classifier
    This classifier supports unlabeled data by overriding forward_train
    """

    def forward_train(self, imgs, **kwargs):
        """Data is transmitted as a classifier training function

        Args:
            imgs (list[Tensor]): List of tensors of shape (1, C, H, W)
                Typically these should be mean centered and std scaled.
            kwargs (keyword arguments): Specific to concrete implementation
        """
        if "gt_label" not in kwargs:
            raise ValueError("'gt_label' does not exist in the labeled image")
        if "extra_0" not in kwargs:
            raise ValueError("'extra_0' does not exist in the dataset")
        target = kwargs["gt_label"].squeeze()
        unlabeled_data = kwargs["extra_0"]
        x = {}
        x["labeled"] = self.extract_feat(imgs)

        img_uw = unlabeled_data["img"]
        # weakly augmented images are used only for getting the pseudo label.
        # not required to calculate gradients.
        with torch.no_grad():
            x["unlabeled_weak"] = self.extract_feat(img_uw)

        img_us = unlabeled_data["img_strong"]
        x["unlabeled_strong"] = self.extract_feat(img_us)

        losses = dict()
        loss = self.head.forward_train(x, target)
        losses.update(loss)

        return losses



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import numpy as np
import pytest

from otx.algorithms.classification.adapters.mmcls.data import OTXClsDataset
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)


def load_test_dataset():
    """Helper to create test dataset"""

    def gen_image(resolution, x1, y1, x2, y2):
        w, h = resolution
        image = np.full([h, w, 3], fill_value=255, dtype=np.uint8)
        image[int(y1 * h) : int(y2 * h), int(x1 * w) : int(x2 * w), :] = np.array([0, 128, 128], dtype=np.uint8)[
            None, None, :
        ]
        return image, Rectangle(x1=x1, y1=y1, x2=x2, y2=y2)

    images = [
        gen_image((640, 480), 0.0, 0.0, 0.5, 0.5),
        gen_image((640, 480), 0.5, 0.0, 1.0, 0.5),
        gen_image((640, 480), 0.0, 0.5, 0.5, 1.0),
        gen_image((640, 480), 0.5, 0.5, 1.0, 1.0),
    ]
    labels = [LabelEntity(name="rect", domain=Domain.DETECTION, id=ID("0"))]

    def get_image(i, subset):
        image, bbox = images[i]
        return DatasetItemEntity(
            media=Image(data=image),
            annotation_scene=AnnotationSceneEntity(
                annotations=[Annotation(bbox, labels=[ScoredLabel(label=labels[0])])],
                kind=AnnotationSceneKind.ANNOTATION,
            ),
            subset=subset,
        )

    items = [
        get_image(0, Subset.TRAINING),
        get_image(1, Subset.TRAINING),
        get_image(2, Subset.TRAINING),
        get_image(3, Subset.TRAINING),
        get_image(0, Subset.TRAINING),
        get_image(1, Subset.TRAINING),
        get_image(2, Subset.TRAINING),
        get_image(3, Subset.TRAINING),
        get_image(0, Subset.TRAINING),
        get_image(1, Subset.TRAINING),
        get_image(0, Subset.VALIDATION),
        get_image(1, Subset.VALIDATION),
        get_image(2, Subset.VALIDATION),
        get_image(3, Subset.VALIDATION),
        get_image(0, Subset.TESTING),
        get_image(1, Subset.TESTING),
        get_image(2, Subset.TESTING),
        get_image(3, Subset.TESTING),
    ]
    return DatasetEntity(items), labels


class TestOTXClsDatasetInputParamsValidation:
    @e2e_pytest_unit
    def test_otx_classification_dataset_init_params_validation(self):
        """
        <b>Description:</b>
        Check OTXClsDataset object initialization parameters validation

        <b>Input data:</b>
        OTXClsDataset object initialization parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        OTXClsDataset initialization parameter
        """
        dataset, labels_list = load_test_dataset()

        correct_values_dict = {
            "otx_dataset": dataset,
            "labels": labels_list,
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "otx_dataset" parameter
            ("otx_dataset", unexpected_str),
            # Unexpected string is specified as "labels" parameter
            ("labels", unexpected_str),
            # Unexpected string is specified as nested label
            ("labels", [labels_list[0], unexpected_str]),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=OTXClsDataset,
        )

    @e2e_pytest_unit
    def test_otx_classification_dataset_getitem_params_validation(self):
        """
        <b>Description:</b>
        Check OTXClsDataset object "__getitem__" method input parameters validation

        <b>Input data:</b>
        "idx" non-integer parameter

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "__getitem__" method
        """
        dataset, labels_list = load_test_dataset()
        otx_classification_dataset = OTXClsDataset(otx_dataset=dataset, labels=labels_list)
        with pytest.raises(ValueError):
            otx_classification_dataset.__getitem__(index="unexpected string")  # type: ignore



"""Datumaro Helper."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=invalid-name
import os
from typing import List, Tuple, Union

import datumaro
from datumaro.components.dataset import Dataset, DatasetSubset
from datumaro.components.dataset_base import DatasetItem
from datumaro.plugins.splitter import Split


class DatasetManager:
    """The aim of DatasetManager is support datumaro functions at easy use.

    All kind of functions implemented in Datumaro are supported by this Manager.
    Since DatasetManager just wraps Datumaro's function,
    All methods are implemented as static method.
    """

    @staticmethod
    def get_train_dataset(dataset: Dataset) -> DatasetSubset:
        """Returns train dataset."""
        for k, v in dataset.subsets().items():
            if "train" in k or "default" in k:
                return v
        raise ValueError("Can't find training data.")

    @staticmethod
    def get_val_dataset(dataset: Dataset) -> Union[DatasetSubset, None]:
        """Returns validation dataset."""
        for k, v in dataset.subsets().items():
            if "val" in k:
                return v
        return None

    @staticmethod
    def get_data_format(data_root: str) -> str:
        """Find the format of dataset."""
        data_root = os.path.abspath(data_root)

        data_format: str = ""

        # TODO #
        # Currently, below `if/else` statements is mandatory
        # because Datumaro can't detect the multi-cvat and mvtec.
        # After, the upgrade of Datumaro, below codes will be changed.
        if DatasetManager.is_cvat_format(data_root):
            data_format = "multi-cvat"
        elif DatasetManager.is_mvtec_format(data_root):
            data_format = "mvtec"
        else:
            data_formats = datumaro.Environment().detect_dataset(data_root)
            # TODO: how to avoid hard-coded part
            data_format = data_formats[0] if "imagenet" not in data_formats else "imagenet"
        print(f"[*] Detected dataset format: {data_format}")
        return data_format

    @staticmethod
    def get_image_path(data_item: DatasetItem) -> str:
        """Returns the path of image."""
        return data_item.media.path

    @staticmethod
    def export_dataset(dataset: Dataset, output_dir: str, data_format: str, save_media=True):
        """Export the Datumaro Dataset."""
        return dataset.export(output_dir, data_format, save_media=save_media)

    @staticmethod
    def import_dataset(data_root: str, data_format: str) -> dict:
        """Import dataset."""
        return Dataset.import_from(data_root, format=data_format)

    @staticmethod
    def auto_split(task: str, dataset: Dataset, split_ratio: List[Tuple[str, float]]) -> dict:
        """Automatically split the dataset: train --> train/val."""
        splitter = Split(dataset, task.lower(), split_ratio)
        return splitter.subsets()

    @staticmethod
    def is_cvat_format(path: str) -> bool:
        """Detect whether data path is CVAT format or not.

        Currently, we used multi-video CVAT format for Action tasks.

        This function can detect the multi-video CVAT format.

        Multi-video CVAT format
        root
        |--video_0
            |--images
                |--frame0001.png
            |--annotations.xml
        |--video_1
        |--video_2

        will be deprecated soon.
        """

        cvat_format = sorted(["images", "annotations.xml"])
        for sub_folder in os.listdir(path):
            # video_0, video_1, ...
            sub_folder_path = os.path.join(path, sub_folder)
            # files must be same with cvat_format
            if os.path.isdir(sub_folder_path):
                files = sorted(os.listdir(sub_folder_path))
                if files != cvat_format:
                    return False
        return True

    @staticmethod
    def is_mvtec_format(path: str) -> bool:
        """Detect whether data path is MVTec format or not.

        Check the first-level architecture folder, to know whether the dataset is MVTec or not.

        MVTec default structure like as below:
        root
        |--ground_truth
        |--train
        |--test

        will be deprecated soon.
        """

        mvtec_format = sorted(["ground_truth", "train", "test"])
        folder_list = []
        for sub_folder in os.listdir(path):
            sub_folder_path = os.path.join(path, sub_folder)
            # only use the folder name.
            if os.path.isdir(sub_folder_path):
                folder_list.append(sub_folder)
        return sorted(folder_list) == mvtec_format



"""Faster RCNN config for Action classification backbone + Fast RCNN model's inference."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

faster_rcnn = dict(
    type="FasterRCNN",
    backbone=dict(
        type="ResNet",
        depth=50,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        frozen_stages=1,
        norm_cfg=dict(type="BN", requires_grad=True),
        norm_eval=True,
        style="pytorch",
        init_cfg=dict(type="Pretrained", checkpoint="torchvision://resnet50"),
    ),
    neck=dict(type="FPN", in_channels=[256, 512, 1024, 2048], out_channels=256, num_outs=5),
    rpn_head=dict(
        type="RPNHead",
        in_channels=256,
        feat_channels=256,
        anchor_generator=dict(type="AnchorGenerator", scales=[8], ratios=[0.5, 1.0, 2.0], strides=[4, 8, 16, 32, 64]),
        bbox_coder=dict(type="DeltaXYWHBBoxCoder", target_means=[0.0, 0.0, 0.0, 0.0], target_stds=[1.0, 1.0, 1.0, 1.0]),
        loss_cls=dict(type="CrossEntropyLoss", use_sigmoid=True, loss_weight=1.0),
        loss_bbox=dict(type="L1Loss", loss_weight=1.0),
    ),
    roi_head=dict(
        type="StandardRoIHead",
        bbox_roi_extractor=dict(
            type="SingleRoIExtractor",
            roi_layer=dict(type="RoIAlign", output_size=7, sampling_ratio=0),
            out_channels=256,
            featmap_strides=[4, 8, 16, 32],
        ),
        bbox_head=dict(
            type="Shared2FCBBoxHead",
            in_channels=256,
            fc_out_channels=1024,
            roi_feat_size=7,
            num_classes=80,
            bbox_coder=dict(
                type="DeltaXYWHBBoxCoder", target_means=[0.0, 0.0, 0.0, 0.0], target_stds=[0.1, 0.1, 0.2, 0.2]
            ),
            reg_class_agnostic=False,
            loss_cls=dict(type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0),
            loss_bbox=dict(type="L1Loss", loss_weight=1.0),
        ),
    ),
    # model training and testing settings
    train_cfg=dict(
        rpn=dict(
            assigner=dict(
                type="MaxIoUAssigner",
                pos_iou_thr=0.7,
                neg_iou_thr=0.3,
                min_pos_iou=0.3,
                match_low_quality=True,
                ignore_iof_thr=-1,
            ),
            sampler=dict(type="RandomSampler", num=256, pos_fraction=0.5, neg_pos_ub=-1, add_gt_as_proposals=False),
            allowed_border=-1,
            pos_weight=-1,
            debug=False,
        ),
        rpn_proposal=dict(nms_pre=2000, max_per_img=1000, nms=dict(type="nms", iou_threshold=0.7), min_bbox_size=0),
        rcnn=dict(
            assigner=dict(
                type="MaxIoUAssigner",
                pos_iou_thr=0.5,
                neg_iou_thr=0.5,
                min_pos_iou=0.5,
                match_low_quality=False,
                ignore_iof_thr=-1,
            ),
            sampler=dict(type="RandomSampler", num=512, pos_fraction=0.25, neg_pos_ub=-1, add_gt_as_proposals=True),
            pos_weight=-1,
            debug=False,
        ),
    ),
    test_cfg=dict(
        rpn=dict(nms_pre=1000, max_per_img=1000, nms=dict(type="nms", iou_threshold=0.7), min_bbox_size=0),
        rcnn=dict(score_thr=0.05, nms=dict(type="nms", iou_threshold=0.5), max_per_img=100)
        # soft-nms is also supported for rcnn testing
        # e.g., nms=dict(type='soft_nms', iou_threshold=0.5, min_score=0.05)
    ),
)

faster_rcnn_pretrained = (
    "http://download.openmmlab.com/mmdetection/v2.0/faster_rcnn/faster_rcnn_r50_fpn_2x_coco/"
    "faster_rcnn_r50_fpn_2x_coco_bbox_mAP-0.384_20200504_210434-a5d8aa15.pth"
)



"""Unit Test for otx.algorithms.detection.adapters.mmdet.utils.config_utils."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import math
import tempfile
from typing import List

import pytest
from mmcv.utils import Config, ConfigDict

from otx.algorithms.common.adapters.mmcv.utils import is_epoch_based_runner
from otx.algorithms.detection.adapters.mmdet.utils.config_utils import (
    cluster_anchors,
    patch_adaptive_repeat_dataset,
    patch_config,
    patch_datasets,
    patch_evaluation,
    patch_model_config,
    prepare_for_training,
    set_hyperparams,
    should_cluster_anchors,
)
from otx.algorithms.detection.configs.base import DetectionConfig
from otx.api.entities.label import Domain
from otx.api.entities.model_template import TaskType
from otx.mpa.utils.config_utils import MPAConfig
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.detection.test_helpers import (
    DEFAULT_DET_MODEL_CONFIG_PATH,
    DEFAULT_ISEG_MODEL_CONFIG_PATH,
    generate_det_dataset,
    generate_labels,
)


@e2e_pytest_unit
@pytest.mark.parametrize("domain", [Domain.DETECTION, Domain.INSTANCE_SEGMENTATION])
def test_patch_config(domain):
    """Test patch_config function."""
    config = Config(
        dict(
            evaluation=dict(),
            checkpoint_config=dict(),
            data=dict(train=dict(labels=dict()), val=dict(labels=dict()), test=dict(labels=dict())),
        )
    )
    labels = generate_labels(1, domain)
    with tempfile.TemporaryDirectory() as work_dir:
        patch_config(config, work_dir, labels)
    assert "train_pipeline" not in config
    assert "test_pipeline" not in config
    assert "gpu_ids" in config
    assert "work_dir" in config
    assert "max_keep_ckpts" in config.checkpoint_config
    assert "interval" in config.checkpoint_config
    assert "labels" in config.data.train
    assert "labels" in config.data.val
    assert "labels" in config.data.test


@e2e_pytest_unit
@pytest.mark.parametrize(
    "cfg_path, domain",
    [
        (DEFAULT_DET_MODEL_CONFIG_PATH, Domain.DETECTION),
        (
            DEFAULT_ISEG_MODEL_CONFIG_PATH,
            Domain.INSTANCE_SEGMENTATION,
        ),
    ],
)
@pytest.mark.parametrize("num_classes", [1, 3, 10, 80])
def test_patch_model_config(num_classes, cfg_path, domain):
    """Test patch_model_config function."""
    config = MPAConfig.fromfile(cfg_path)
    labels = generate_labels(num_classes, domain)
    head_names = ("mask_head", "bbox_head", "segm_head")
    patch_model_config(config, labels)
    if "roi_head" in config.model:
        for head_name in head_names:
            if head_name in config.model.roi_head:
                if isinstance(config.model.roi_head[head_name], List):
                    for head in config.model.roi_head[head_name]:
                        assert head.num_classes == 1
                else:
                    assert config.model.roi_head[head_name].num_classes == num_classes
    else:
        for head_name in head_names:
            if head_name in config.model:
                assert config.model[head_name].num_classes == num_classes


@e2e_pytest_unit
def test_set_hyperparams():
    """Test set_hyperparams function."""
    config = Config(dict(data=dict(), optimizer=dict(), lr_config=dict(), runner=dict(type="EpochBasedRunner")))
    hyperparams = DetectionConfig()
    set_hyperparams(config, hyperparams)
    assert config.data.samples_per_gpu == int(hyperparams.learning_parameters.batch_size)
    assert config.data.workers_per_gpu == int(hyperparams.learning_parameters.num_workers)
    assert config.optimizer.lr == float(hyperparams.learning_parameters.learning_rate)
    assert config.lr_config.warmup_iters == int(hyperparams.learning_parameters.learning_rate_warmup_iters)
    total_iterations = int(hyperparams.learning_parameters.num_iters)
    if is_epoch_based_runner(config.runner):
        assert config.runner.max_epochs == total_iterations
    else:
        assert config.runner.max_iters == total_iterations


@e2e_pytest_unit
@pytest.mark.parametrize("num_samples", [10, 10000])
@pytest.mark.parametrize("decay", [-0.002, -0.2])
@pytest.mark.parametrize("factor", [30, 3000])
@pytest.mark.parametrize("dataset_type", ["MultiImageMixDataset", "RepeatDataset"])
def test_patch_adaptive_repeat_dataset(num_samples, decay, factor, dataset_type):
    """Test patch_adaptive_repeat function."""
    cur_epoch, cur_repeat = 2, 1
    config = Config(
        dict(
            runner=dict(type="EpochBasedRunner", max_epochs=cur_epoch),
            data=dict(train=dict(type=dataset_type, adaptive_repeat_times=True, times=cur_repeat)),
        )
    )
    if dataset_type == "MultiImageMixDataset":
        config.merge_from_dict(
            dict(
                data=dict(train=dict(dataset=dict(type="RepeatDataset", adaptive_repeat_times=True, times=cur_repeat)))
            )
        )
    patch_adaptive_repeat_dataset(config, num_samples=num_samples, decay=decay, factor=factor)
    adaptive_repeat = max(round(math.exp(decay * num_samples) * factor), 1)
    adaptive_epoch = math.ceil(cur_epoch / num_samples)
    if adaptive_epoch == 1:
        assert config.data.train.times == cur_repeat
        assert config.runner.max_epochs == cur_epoch
    else:
        assert config.data.train.times == adaptive_repeat
        assert config.runner.max_epochs == adaptive_epoch


@e2e_pytest_unit
def test_prepare_for_training():
    """Test prepare_for_training function."""
    with tempfile.TemporaryDirectory() as work_dir:
        config = Config(
            dict(
                data=dict(
                    train=dict(type="RepeatDataset", adaptive_repeat_times=True, otx_dataset=[]),
                    val=dict(),
                    test=dict(),
                ),
                work_dir=work_dir,
                runner=dict(type="EpochBasedRunner", max_epochs=1),
            )
        )
        data_config = ConfigDict(data=dict(train=dict(otx_dataset=[1, 2, 3]), val=dict(), test=dict()))
        prepare_for_training(config, data_config)
        assert "meta" in config.runner
        assert "checkpoints_round_0" in config.work_dir
        assert config.data.train.otx_dataset == [1, 2, 3]


@e2e_pytest_unit
def test_patch_datasets():
    """Test patch_datasets function."""
    config = Config(
        dict(
            data=dict(train=dict(), val=dict(), test=dict(), unlabeled=dict()),
        )
    )
    patch_datasets(config, type="FakeType")
    data_cfg = config.data
    assert "domain" in data_cfg.train
    assert "domain" in data_cfg.val
    assert "domain" in data_cfg.test
    assert "domain" in data_cfg.unlabeled
    assert "otx_dataset" in data_cfg.train
    assert "otx_dataset" in data_cfg.val
    assert "otx_dataset" in data_cfg.test
    assert "otx_dataset" in data_cfg.unlabeled
    assert "labels" in data_cfg.train
    assert "labels" in data_cfg.val
    assert "labels" in data_cfg.test
    assert "labels" in data_cfg.unlabeled
    assert "train_dataloader" in data_cfg
    assert "val_dataloader" in data_cfg
    assert "test_dataloader" in data_cfg


@e2e_pytest_unit
def test_patch_evaluation():
    """Test patch_evaluation function."""
    config = Config(dict(evaluation=dict()))
    patch_evaluation(config)
    assert "metric" in config.evaluation
    assert "save_best" in config.evaluation
    assert "early_stop_metric" in config


@e2e_pytest_unit
@pytest.mark.parametrize("reclustering_anchors", [True, False])
def test_should_cluster_anchors(reclustering_anchors):
    """Test should_cluster_anchors function."""
    config = Config(
        dict(
            model=dict(
                bbox_head=dict(
                    anchor_generator=dict(reclustering_anchors=reclustering_anchors),
                )
            )
        )
    )
    out = should_cluster_anchors(config)
    assert out == reclustering_anchors


@e2e_pytest_unit
@pytest.mark.parametrize("widths, heights", [([1, 10], [1, 10]), ([1, 3, 5, 7, 9], [1, 3, 5, 7, 9])])
def test_cluster_anchors(widths, heights):
    """Test cluster_anchors function."""
    model_config = Config(dict(model=dict(bbox_head=dict(anchor_generator=dict(widths=[widths], heights=[heights])))))

    data_config = Config(
        dict(
            data=dict(
                test=dict(
                    pipeline=[
                        dict(type="LoadImageFromFile"),
                        dict(
                            type="MultiScaleFlipAug",
                            img_scale=(10, 10),
                        ),
                    ]
                )
            )
        )
    )
    dataset, _ = generate_det_dataset(task_type=TaskType.DETECTION)
    cluster_anchors(model_config, data_config, dataset)



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


import pytest

from otx.algorithms.anomaly.adapters.anomalib.config import get_anomalib_config
from otx.algorithms.anomaly.configs.classification.draem import (
    DraemAnomalyClassificationConfig,
)
from otx.algorithms.anomaly.configs.classification.padim import (
    PadimAnomalyClassificationConfig,
)
from otx.algorithms.anomaly.configs.classification.stfpm import (
    STFPMAnomalyClassificationConfig,
)
from otx.algorithms.anomaly.configs.detection.draem import DraemAnomalyDetectionConfig
from otx.algorithms.anomaly.configs.detection.padim import PadimAnomalyDetectionConfig
from otx.algorithms.anomaly.configs.detection.stfpm import STFPMAnomalyDetectionConfig
from otx.algorithms.anomaly.configs.segmentation.draem import (
    DraemAnomalySegmentationConfig,
)
from otx.algorithms.anomaly.configs.segmentation.padim import (
    PadimAnomalySegmentationConfig,
)
from otx.algorithms.anomaly.configs.segmentation.stfpm import (
    STFPMAnomalySegmentationConfig,
)
from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.configuration.helper import convert, create
from tests.unit.algorithms.anomaly.helpers.utils import get_model_template


@pytest.mark.parametrize(
    ["model_name", "configurable_parameters"],
    [
        ("padim", PadimAnomalyClassificationConfig),
        ("padim", PadimAnomalyDetectionConfig),
        ("padim", PadimAnomalySegmentationConfig),
        ("stfpm", STFPMAnomalyClassificationConfig),
        ("stfpm", STFPMAnomalyDetectionConfig),
        ("stfpm", STFPMAnomalySegmentationConfig),
        ("draem", DraemAnomalyClassificationConfig),
        ("draem", DraemAnomalyDetectionConfig),
        ("draem", DraemAnomalySegmentationConfig),
    ],
)
def test_model_template_loading(model_name, configurable_parameters):
    """Test that we can load the model template and create a config from it."""
    # Create from class
    configuration = configurable_parameters()
    configurable_parameters_yaml_str = convert(configuration, str)
    assert isinstance(configurable_parameters_yaml_str, str)

    # Check if it can be created from yaml
    configurable_parameters_yaml: ConfigurableParameters = create(configurable_parameters_yaml_str)

    model_template = get_model_template(model_name)
    hyper_parameters: dict = model_template.hyper_parameters.data
    configurable_parameters_loaded = create(hyper_parameters)

    assert configurable_parameters_yaml == configurable_parameters_loaded

    # Confirm that we can create an anomalib config from the loaded yaml
    get_anomalib_config(model_name, configurable_parameters_loaded)



"""Setup file for OTX."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import os
import subprocess
import sys
import warnings
from collections import defaultdict
from glob import glob
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from typing import List, Optional, Union

import Cython
import numpy
from Cython.Build import cythonize
from pkg_resources import Requirement
from setuptools import Extension, find_packages, setup

try:
    from torch.utils.cpp_extension import BuildExtension

    cmd_class = {"build_ext": BuildExtension}
except ModuleNotFoundError:
    cmd_class = {}
    print("Skip building ext ops due to the absence of torch.")


def readme():
    with open('README.md', encoding='utf-8') as f:
        content = f.read()
    return content


def load_module(name: str = "otx/__init__.py"):
    """Load Python Module.

    Args:
        name (str, optional): Name of the module to load.
            Defaults to "otx/__init__.py".
    """
    location = str(Path(__file__).parent / name)
    spec = spec_from_file_location(name=name, location=location)
    module = module_from_spec(spec)  # type: ignore
    spec.loader.exec_module(module)  # type: ignore
    return module


def get_otx_version() -> str:
    """Get version from `otx.__init__`.

    Version is stored in the main __init__ module in `otx`.
    The varible storing the version is `__version__`. This function
    reads `__init__` file, checks `__version__ variable and return
    the value assigned to it.

    Example:
        >>> # Assume that __version__ = "0.2.6"
        >>> get_otx_version()
        "0.2.6"

    Returns:
        str: `otx` version.
    """
    otx = load_module(name="otx/__init__.py")
    return otx.__version__


def get_requirements(requirement_files: Union[str, List[str]]) -> List[str]:
    """Get packages from requirements.txt file.

    This function returns list of required packages from requirement files.

    Args:
        requirement_files (List[str]): txt files that contains list of required
            packages.

    Example:
        >>> get_requirements(requirement_files=["openvino"])
        ['onnx>=1.8.1', 'networkx~=2.5', 'openvino-dev==2021.4.1', ...]

    Returns:
        List[str]: List of required packages
    """
    if isinstance(requirement_files, str):
        requirement_files = [requirement_files]

    requirements: List[str] = []
    for requirement_file in requirement_files:
        with open(f"requirements/{requirement_file}.txt", "r", encoding="UTF-8") as file:
            for line in file:
                package = line.strip()
                if package and not package.startswith(("#", "-f")):
                    requirement = Requirement.parse(package)
                    requirements.append(package)

    return requirements


def get_extensions():
    def _cython_modules():
        package_root = os.path.dirname(__file__)

        cython_files = [
            "otx/mpa/modules/datasets/pipelines/transforms/cython_augments/pil_augment.pyx",
            "otx/mpa/modules/datasets/pipelines/transforms/cython_augments/cv_augment.pyx",
        ]

        ext_modules = [
            Extension(
                cython_file.rstrip(".pyx").replace("/", "."),
                [os.path.join(package_root, cython_file)],
                include_dirs=[numpy.get_include()],
                extra_compile_args=["-O3"],
            )
            for cython_file in cython_files
        ]

        return cythonize(ext_modules, annotate=True)

    extensions = []
    extensions.extend(_cython_modules())
    return extensions


REQUIRED_PACKAGES = get_requirements(requirement_files="api")
EXTRAS_REQUIRE = {
    "action": get_requirements(requirement_files=[
            "base", "openvino", "action",
        ]
    ),
    "anomaly": get_requirements(requirement_files=[
            "base", "openvino", "anomaly",
        ]
    ),
    "classification": get_requirements(requirement_files=[
            "base", "openvino", "classification",
        ]
    ),
    "detection": get_requirements(requirement_files=[
            "base", "openvino", "detection",
        ]
    ),
    "segmentation": get_requirements(requirement_files=[
            "base", "openvino", "segmentation",
        ]
    ),
    "full": get_requirements(requirement_files=[
            "base",
            "openvino",
            "anomaly",
            "classification",
            "detection",
            "segmentation",
            "action",
        ]
    ),
}


def find_yaml_recipes():
    """Find YAML recipe files in the package."""
    results = defaultdict(list)

    for root, _, files in os.walk("otx"):
        module = ".".join(root.split(os.sep))
        for file in files:
            _, ext = os.path.splitext(file)
            if ext in [".yaml", ".json"]:
                results[module] += [file]

    return results


package_data = {"": ["requirements.txt", "README.md", "LICENSE"]}  # Needed for exportable code
package_data.update(find_yaml_recipes())

setup(
    name="otx",
    version=get_otx_version(),
    description="OpenVINO™ Training Extensions: "
    "Train, Evaluate, Optimize, Deploy Computer Vision Models via OpenVINO™",
    long_description=readme(),
    long_description_content_type="text/markdown",
    author="OpenVINO™ Training Extensions Contributors",
    url="https://github.com/openvinotoolkit/training_extensions",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Cython",
    ],
    license="Apache License 2.0",
    packages=find_packages(exclude=("tests",)),
    package_data=package_data,
    ext_modules=get_extensions(),
    cmdclass=cmd_class,
    install_requires=REQUIRED_PACKAGES,
    extras_require=EXTRAS_REQUIRE,
    entry_points={
        "console_scripts": [
            "otx=otx.cli.tools.cli:main",
            "otx_demo=otx.cli.tools.demo:main",
            "otx_eval=otx.cli.tools.eval:main",
            "otx_export=otx.cli.tools.export:main",
            "otx_find=otx.cli.tools.find:main",
            "otx_train=otx.cli.tools.train:main",
            "otx_optimize=otx.cli.tools.optimize:main",
            "otx_build=otx.cli.tools.build:main",
        ]
    },
)



import pytest

from otx.mpa.seg.incremental.stage import IncrSegStage
from otx.mpa.utils.config_utils import MPAConfig
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.segmentation.test_helpers import DEFAULT_RECIPE_CONFIG_PATH


class TestOTXSegStage:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        cfg = MPAConfig.fromfile(DEFAULT_RECIPE_CONFIG_PATH)
        self.stage = IncrSegStage(name="", mode="train", config=cfg, common_cfg=None, index=0)

    @e2e_pytest_unit
    def test_configure_task(self, mocker):
        mock_update_hook = mocker.patch("otx.mpa.seg.incremental.stage.update_or_add_custom_hook")
        self.stage.configure_task(self.stage.cfg, True)

        mock_update_hook.assert_called_once()



"""Converters for output of inferencers."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc
from typing import Any, Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
from openvino.model_zoo.model_api.models import utils

from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.id import ID
from otx.api.entities.label import Domain
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.utils.anomaly_utils import create_detection_annotation_from_anomaly_heatmap
from otx.api.utils.labels_utils import get_empty_label
from otx.api.utils.segmentation_utils import create_annotation_from_segmentation_map
from otx.api.utils.time_utils import now


class IPredictionToAnnotationConverter(metaclass=abc.ABCMeta):
    """Interface for converter."""

    @abc.abstractmethod
    def convert_to_annotation(self, predictions: Any, metadata: Dict) -> AnnotationSceneEntity:
        """Convert raw predictions to AnnotationScene format.

        Args:
            predictions (Any): raw predictions from inferencer
            metadata (Dict): metadata from inferencer

        Returns:
            AnnotationSceneEntity: annotation object containing the shapes obtained from the raw predictions.
        """
        raise NotImplementedError


class DetectionToAnnotationConverter(IPredictionToAnnotationConverter):
    """Converts Object Detections to Annotations.

    Args:
        labels (List[LabelEntity]): list of labels
    """

    def __init__(self, labels: Union[LabelSchemaEntity, List]):
        self.labels = labels.get_labels(include_empty=False) if isinstance(labels, LabelSchemaEntity) else labels
        self.label_map = dict(enumerate(self.labels))

    def convert_to_annotation(
        self, predictions: np.ndarray, metadata: Optional[Dict[str, np.ndarray]] = None
    ) -> AnnotationSceneEntity:
        """Convert predictions to annotation format.

        Args:
            predictions (np.ndarray): Prediction with shape [num_predictions, 6] or
                            [num_predictions, 7]
                Supported detection formats are

                * [label, confidence, x1, y1, x2, y2]
                * [_, label, confidence, x1, y1, x2, y2]

                .. note::
                `label` can be any integer that can be mapped to `self.labels`
                `confidence` should be a value between 0 and 1
                `x1`, `x2`, `y1` and `y2` are expected to be normalized.
            metadata (Optional[Dict]): Additional information

        Returns:
            AnnotationScene: AnnotationScene Object containing the boxes obtained from the prediction.
        """
        if metadata:
            predictions[:, 2:] /= np.tile(metadata["original_shape"][1::-1], 2)
        annotations = self.__convert_to_annotations(predictions)
        # media_identifier = ImageIdentifier(image_id=ID())
        annotation_scene = AnnotationSceneEntity(
            id=ID(),
            kind=AnnotationSceneKind.PREDICTION,
            editor="otx",
            creation_date=now(),
            annotations=annotations,
        )

        return annotation_scene

    def __convert_to_annotations(self, predictions: np.ndarray) -> List[Annotation]:
        """Converts a list of Detections to OTX Annotation objects.

        Args:
            predictions (np.ndarray): A list of predictions with shape [num_prediction, 6] or
                            [num_predictions, 7]

        Returns:
            List[Annotation]: A list of Annotation objects with Rectangle shapes

        Raises:
            ValueError: This error is raised if the shape of prediction is not
                            (n, 7) or (n, 6)
        """
        annotations = []
        if len(predictions) and predictions.shape[1:] < (6,) or predictions.shape[1:] > (7,):
            raise ValueError(
                f"Shape of prediction is not expected, expected (n, 7) or (n, 6) but got {predictions.shape}"
            )

        for prediction in predictions:
            if prediction.shape == (7,):
                # Some OpenVINO models use an output shape of [7,]
                # If this is the case, skip the first value as it is not used
                prediction = prediction[1:]

            label = int(prediction[0])
            confidence = prediction[1]
            scored_label = ScoredLabel(self.label_map[label], confidence)
            coords = prediction[2:]
            annotations.append(
                Annotation(
                    Rectangle(coords[0], coords[1], coords[2], coords[3]),
                    labels=[scored_label],
                )
            )

        return annotations


def create_converter(
    converter_type: Domain,
    labels: LabelSchemaEntity,
) -> IPredictionToAnnotationConverter:
    """Simple factory for converters based on type of tasks.

    Args:
        converter_type (Domain): type of converter
        labels (LabelSchemaEntity): label schema entity
    """

    converter: IPredictionToAnnotationConverter
    if converter_type == Domain.DETECTION:
        converter = DetectionToAnnotationConverter(labels)
    elif converter_type == Domain.SEGMENTATION:
        converter = SegmentationToAnnotationConverter(labels)
    elif converter_type == Domain.CLASSIFICATION:
        converter = ClassificationToAnnotationConverter(labels)
    elif converter_type == Domain.ANOMALY_CLASSIFICATION:
        converter = AnomalyClassificationToAnnotationConverter(labels)
    elif converter_type == Domain.ANOMALY_DETECTION:
        converter = AnomalyDetectionToAnnotationConverter(labels)
    elif converter_type == Domain.ANOMALY_SEGMENTATION:
        converter = AnomalySegmentationToAnnotationConverter(labels)
    elif converter_type == Domain.INSTANCE_SEGMENTATION:
        converter = MaskToAnnotationConverter(labels)
    elif converter_type == Domain.ROTATED_DETECTION:
        converter = RotatedRectToAnnotationConverter(labels)
    else:
        raise ValueError(f"Unknown converter type: {converter_type}")

    return converter


class DetectionBoxToAnnotationConverter(IPredictionToAnnotationConverter):
    """Converts DetectionBox Predictions ModelAPI to Annotations.

    Args:
        labels (LabelSchemaEntity): Label Schema containing the label info of the task
    """

    def __init__(self, labels: LabelSchemaEntity):
        self.labels = labels.get_labels(include_empty=False)

    def convert_to_annotation(
        self, predictions: List[utils.Detection], metadata: Dict[str, Any]
    ) -> AnnotationSceneEntity:
        """Convert predictions to OTX Annotation Scene using the metadata.

        Args:
            predictions (tuple): Raw predictions from the model.
            metadata (Dict[str, Any]): Variable containing metadata information.

        Returns:
            AnnotationSceneEntity: OTX annotation scene entity object.
        """
        annotations = []
        image_size = metadata["original_shape"][1::-1]
        for box in predictions:
            scored_label = ScoredLabel(self.labels[int(box.id)], float(box.score))
            coords = np.array(box.get_coords(), dtype=float)
            if (coords[2] - coords[0]) * (coords[3] - coords[1]) < 1.0:
                continue
            coords /= np.tile(image_size, 2)
            annotations.append(
                Annotation(
                    Rectangle(coords[0], coords[1], coords[2], coords[3]),
                    labels=[scored_label],
                )
            )

        annotation_scene = AnnotationSceneEntity(
            kind=AnnotationSceneKind.PREDICTION,
            annotations=annotations,
        )
        return annotation_scene


class SegmentationToAnnotationConverter(IPredictionToAnnotationConverter):
    """Converts Segmentation Predictions ModelAPI to Annotations.

    Args:
        labels (LabelSchemaEntity): Label Schema containing the label info of the task
    """

    def __init__(self, label_schema: LabelSchemaEntity):
        labels = label_schema.get_labels(include_empty=False)
        self.label_map = dict(enumerate(labels, 1))

    def convert_to_annotation(self, predictions: np.ndarray, metadata: Dict[str, Any]) -> AnnotationSceneEntity:
        """Convert predictions to OTX Annotation Scene using the metadata.

        Args:
            predictions (tuple): Raw predictions from the model.
            metadata (Dict[str, Any]): Variable containing metadata information.

        Returns:
            AnnotationSceneEntity: OTX annotation scene entity object.
        """
        soft_predictions = metadata.get("soft_predictions", np.ones(predictions.shape))
        annotations = create_annotation_from_segmentation_map(
            hard_prediction=predictions,
            soft_prediction=soft_predictions,
            label_map=self.label_map,
        )

        return AnnotationSceneEntity(kind=AnnotationSceneKind.PREDICTION, annotations=annotations)


class ClassificationToAnnotationConverter(IPredictionToAnnotationConverter):
    """Converts Classification Predictions ModelAPI to Annotations.

    Args:
        labels (LabelSchemaEntity): Label Schema containing the label info of the task
    """

    def __init__(self, label_schema: LabelSchemaEntity):
        if len(label_schema.get_labels(False)) == 1:
            self.labels = label_schema.get_labels(include_empty=True)
        else:
            self.labels = label_schema.get_labels(include_empty=False)
        self.empty_label = get_empty_label(label_schema)
        multilabel = len(label_schema.get_groups(False)) > 1
        multilabel = multilabel and len(label_schema.get_groups(False)) == len(
            label_schema.get_labels(include_empty=False)
        )
        self.hierarchical = not multilabel and len(label_schema.get_groups(False)) > 1

        self.label_schema = label_schema

    def convert_to_annotation(
        self, predictions: List[Tuple[int, float]], metadata: Optional[Dict] = None
    ) -> AnnotationSceneEntity:
        """Convert predictions to OTX Annotation Scene using the metadata.

        Args:
            predictions (tuple): Raw predictions from the model.
            metadata (Dict[str, Any]): Variable containing metadata information.

        Returns:
            AnnotationSceneEntity: OTX annotation scene entity object.
        """
        labels = []
        for index, score in predictions:
            labels.append(ScoredLabel(self.labels[index], float(score)))
        if self.hierarchical:
            labels = self.label_schema.resolve_labels_probabilistic(labels)

        if not labels and self.empty_label:
            labels = [ScoredLabel(self.empty_label, probability=1.0)]

        annotations = [Annotation(Rectangle.generate_full_box(), labels=labels)]
        return AnnotationSceneEntity(kind=AnnotationSceneKind.PREDICTION, annotations=annotations)


class AnomalyClassificationToAnnotationConverter(IPredictionToAnnotationConverter):
    """Converts AnomalyClassification Predictions ModelAPI to Annotations.

    Args:
        labels (LabelSchemaEntity): Label Schema containing the label info of the task
    """

    def __init__(self, label_schema: LabelSchemaEntity):
        labels = label_schema.get_labels(include_empty=False)
        self.normal_label = [label for label in labels if not label.is_anomalous][0]
        self.anomalous_label = [label for label in labels if label.is_anomalous][0]

    def convert_to_annotation(self, predictions: np.ndarray, metadata: Dict[str, Any]) -> AnnotationSceneEntity:
        """Convert predictions to OTX Annotation Scene using the metadata.

        Args:
            predictions (tuple): Raw predictions from the model.
            metadata (Dict[str, Any]): Variable containing metadata information.

        Returns:
            AnnotationSceneEntity: OTX annotation scene entity object.
        """
        pred_label = predictions >= metadata.get("threshold", 0.5)

        label = self.anomalous_label if pred_label else self.normal_label
        probability = (1 - predictions) if predictions < 0.5 else predictions

        annotations = [
            Annotation(
                Rectangle.generate_full_box(),
                labels=[ScoredLabel(label=label, probability=float(probability))],
            )
        ]
        return AnnotationSceneEntity(kind=AnnotationSceneKind.PREDICTION, annotations=annotations)


class AnomalySegmentationToAnnotationConverter(IPredictionToAnnotationConverter):
    """Converts AnomalyClassification Predictions ModelAPI to Annotations.

    Args:
        labels (LabelSchemaEntity): Label Schema containing the label info of the task
    """

    def __init__(self, label_schema: LabelSchemaEntity):
        labels = label_schema.get_labels(include_empty=False)
        self.normal_label = [label for label in labels if not label.is_anomalous][0]
        self.anomalous_label = [label for label in labels if label.is_anomalous][0]
        self.label_map = {0: self.normal_label, 1: self.anomalous_label}

    def convert_to_annotation(self, predictions: np.ndarray, metadata: Dict[str, Any]) -> AnnotationSceneEntity:
        """Convert predictions to OTX Annotation Scene using the metadata.

        Args:
            predictions (tuple): Raw predictions from the model.
            metadata (Dict[str, Any]): Variable containing metadata information.

        Returns:
            AnnotationSceneEntity: OTX annotation scene entity object.
        """
        pred_mask = predictions >= 0.5
        mask = pred_mask.squeeze().astype(np.uint8)
        annotations = create_annotation_from_segmentation_map(mask, predictions, self.label_map)
        if len(annotations) == 0:
            # TODO: add confidence to this label
            annotations = [
                Annotation(
                    Rectangle.generate_full_box(),
                    labels=[ScoredLabel(label=self.normal_label, probability=1.0)],
                )
            ]
        return AnnotationSceneEntity(kind=AnnotationSceneKind.PREDICTION, annotations=annotations)


class AnomalyDetectionToAnnotationConverter(IPredictionToAnnotationConverter):
    """Converts Anomaly Detection Predictions ModelAPI to Annotations.

    Args:
        labels (LabelSchemaEntity): Label Schema containing the label info of the task
    """

    def __init__(self, label_schema: LabelSchemaEntity):
        """Initialize AnomalyDetectionToAnnotationConverter.

        Args:
            label_schema (LabelSchemaEntity): Label Schema containing the label info of the task
        """
        labels = label_schema.get_labels(include_empty=False)
        self.normal_label = [label for label in labels if not label.is_anomalous][0]
        self.anomalous_label = [label for label in labels if label.is_anomalous][0]
        self.label_map = {0: self.normal_label, 1: self.anomalous_label}

    def convert_to_annotation(self, predictions: np.ndarray, metadata: Dict[str, Any]) -> AnnotationSceneEntity:
        """Convert predictions to OTX Annotation Scene using the metadata.

        Args:
            predictions (tuple): Raw predictions from the model.
            metadata (Dict[str, Any]): Variable containing metadata information.

        Returns:
            AnnotationSceneEntity: OTX annotation scene entity object.
        """
        pred_mask = predictions >= 0.5
        mask = pred_mask.squeeze().astype(np.uint8)
        annotations = create_detection_annotation_from_anomaly_heatmap(mask, predictions, self.label_map)
        if len(annotations) == 0:
            # TODO: add confidence to this label
            annotations = [
                Annotation(
                    Rectangle.generate_full_box(),
                    labels=[ScoredLabel(label=self.normal_label, probability=1.0)],
                )
            ]
        return AnnotationSceneEntity(kind=AnnotationSceneKind.PREDICTION, annotations=annotations)


class MaskToAnnotationConverter(IPredictionToAnnotationConverter):
    """Converts DetectionBox Predictions ModelAPI to Annotations."""

    def __init__(self, labels: LabelSchemaEntity):
        self.labels = labels.get_labels(include_empty=False)

    def convert_to_annotation(self, predictions: tuple, metadata: Dict[str, Any]) -> AnnotationSceneEntity:
        """Convert predictions to OTX Annotation Scene using the metadata.

        Args:
            predictions (tuple): Raw predictions from the model.
            metadata (Dict[str, Any]): Variable containing metadata information.

        Returns:
            AnnotationSceneEntity: OTX annotation scene entity object.
        """
        annotations = []
        for score, class_idx, _, mask in zip(*predictions):
            mask = mask.astype(np.uint8)
            contours, hierarchies = cv2.findContours(mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)
            if hierarchies is None:
                continue
            for contour, hierarchy in zip(contours, hierarchies[0]):
                if hierarchy[3] != -1:
                    continue
                if len(contour) <= 2 or cv2.contourArea(contour) < 1.0:
                    continue
                contour = list(contour)
                points = [
                    Point(
                        x=point[0][0] / metadata["original_shape"][1],
                        y=point[0][1] / metadata["original_shape"][0],
                    )
                    for point in contour
                ]
                polygon = Polygon(points=points)
                annotations.append(
                    Annotation(
                        polygon,
                        labels=[ScoredLabel(self.labels[int(class_idx) - 1], float(score))],
                    )
                )
        annotation_scene = AnnotationSceneEntity(
            kind=AnnotationSceneKind.PREDICTION,
            annotations=annotations,
        )
        return annotation_scene


class RotatedRectToAnnotationConverter(IPredictionToAnnotationConverter):
    """Converts Rotated Rect (mask) Predictions ModelAPI to Annotations.

    Args:
        labels (LabelSchemaEntity): Label Schema containing the label info of the task
    """

    def __init__(self, labels: LabelSchemaEntity):
        self.labels = labels.get_labels(include_empty=False)

    def convert_to_annotation(self, predictions: tuple, metadata: Dict[str, Any]) -> AnnotationSceneEntity:
        """Convert predictions to OTX Annotation Scene using the metadata.

        Args:
            predictions (tuple): Raw predictions from the model.
            metadata (Dict[str, Any]): Variable containing metadata information.

        Returns:
            AnnotationSceneEntity: OTX annotation scene entity object.
        """
        annotations = []
        for score, class_idx, _, mask in zip(*predictions):
            mask = mask.astype(np.uint8)
            contours, hierarchies = cv2.findContours(mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)
            if hierarchies is None:
                continue
            for contour, hierarchy in zip(contours, hierarchies[0]):
                if hierarchy[3] != -1:
                    continue
                if len(contour) <= 2 or cv2.contourArea(contour) < 1.0:
                    continue
                points = [
                    Point(
                        x=point[0] / metadata["original_shape"][1],
                        y=point[1] / metadata["original_shape"][0],
                    )
                    for point in cv2.boxPoints(cv2.minAreaRect(contour))
                ]
                polygon = Polygon(points=points)
                annotations.append(
                    Annotation(
                        polygon,
                        labels=[ScoredLabel(self.labels[int(class_idx) - 1], float(score))],
                    )
                )
        annotation_scene = AnnotationSceneEntity(
            kind=AnnotationSceneKind.PREDICTION,
            annotations=annotations,
        )
        return annotation_scene



"""Collection of utils about labels in Classifation Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=too-many-nested-blocks, invalid-name

from operator import itemgetter

from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.utils.argument_checks import check_input_parameters_type


@check_input_parameters_type()
def get_multihead_class_info(label_schema: LabelSchemaEntity):  # pylint: disable=too-many-locals
    """Get multihead info by label schema."""
    all_groups = label_schema.get_groups(include_empty=False)
    all_groups_str = []
    for g in all_groups:
        group_labels_str = [lbl.name for lbl in g.labels]
        all_groups_str.append(group_labels_str)

    single_label_groups = [g for g in all_groups_str if len(g) == 1]
    exclusive_groups = [sorted(g) for g in all_groups_str if len(g) > 1]
    single_label_groups.sort(key=itemgetter(0))
    exclusive_groups.sort(key=itemgetter(0))
    class_to_idx = {}
    head_idx_to_logits_range = {}
    num_single_label_classes = 0
    last_logits_pos = 0
    for i, group in enumerate(exclusive_groups):
        head_idx_to_logits_range[i] = (last_logits_pos, last_logits_pos + len(group))
        last_logits_pos += len(group)
        for j, c in enumerate(group):
            class_to_idx[c] = (i, j)  # group idx and idx inside group
            num_single_label_classes += 1

    # other labels are in multilabel group
    for j, group in enumerate(single_label_groups):
        class_to_idx[group[0]] = (len(exclusive_groups), j)

    all_labels = label_schema.get_labels(include_empty=False)
    label_to_idx = {lbl.name: i for i, lbl in enumerate(all_labels)}

    mixed_cls_heads_info = {
        "num_multiclass_heads": len(exclusive_groups),
        "num_multilabel_classes": len(single_label_groups),
        "head_idx_to_logits_range": head_idx_to_logits_range,
        "num_single_label_classes": num_single_label_classes,
        "class_to_group_idx": class_to_idx,
        "all_groups": exclusive_groups + single_label_groups,
        "label_to_idx": label_to_idx,
    }
    return mixed_cls_heads_info



"""Visualisation module."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


from typing import List, Tuple
from warnings import warn

import cv2
import numpy as np
from cv2 import Mat

from otx.api.entities.annotation import Annotation
from otx.api.entities.model_template import TaskType
from otx.api.entities.shapes.polygon import Polygon
from otx.api.entities.shapes.rectangle import Rectangle


def put_text_on_rect_bg(frame: Mat, message: str, position: Tuple[int, int], color=(255, 255, 0)):
    """Puts a text message on a black rectangular aread in specified position of a frame."""

    font_face = cv2.FONT_HERSHEY_COMPLEX
    font_scale = 1
    thickness = 1
    color_bg = (0, 0, 0)
    x, y = position
    text_size, _ = cv2.getTextSize(message, font_face, font_scale, thickness)
    text_w, text_h = text_size
    cv2.rectangle(frame, position, (x + text_w + 1, y + text_h + 1), color_bg, -1)
    cv2.putText(
        frame,
        message,
        (x, y + text_h + font_scale - 1),
        font_face,
        font_scale,
        color,
        thickness,
    )
    return text_size


def draw_masks(frame: Mat, predictions, put_object_count: bool = False):
    """Converts predictions to masks and draw them on frame."""

    frame = frame.copy()
    height, width = frame.shape[0], frame.shape[1]
    segments_image = frame.copy()
    aggregated_mask = np.zeros(frame.shape[:2], dtype=np.uint8)
    aggregated_colored_mask = np.zeros(frame.shape, dtype=np.uint8)
    for prediction in predictions:
        if not isinstance(prediction.shape, Polygon):
            continue
        contours = np.array([[(int(p.x * width), int(p.y * height)) for p in prediction.shape.points]])
        assert len(prediction.get_labels()) == 1
        label = prediction.get_labels()[0]
        color = tuple(getattr(label.color, x) for x in ("blue", "green", "red"))
        mask = np.zeros(shape=(height, width), dtype=np.uint8)
        cv2.drawContours(mask, contours, -1, 255, -1)
        cv2.drawContours(frame, contours, -1, color, 1)
        rect = cv2.boundingRect(contours[0])
        cv2.rectangle(frame, (rect[0], rect[1]), (rect[0] + rect[2], rect[1] + rect[3]), color, 1)
        put_text_on_rect_bg(frame, label.name, (rect[0], rect[1]), color=color)
        cv2.bitwise_or(aggregated_mask, mask, dst=aggregated_mask)
        cv2.bitwise_or(
            aggregated_colored_mask,
            np.asarray(color, dtype=np.uint8),
            dst=aggregated_colored_mask,
            mask=mask,
        )
    # Fill the area occupied by all instances with a colored instances mask image.
    cv2.bitwise_and(
        segments_image,
        np.zeros(3, dtype=np.uint8),
        dst=segments_image,
        mask=aggregated_mask,
    )
    cv2.bitwise_or(
        segments_image,
        aggregated_colored_mask,
        dst=segments_image,
        mask=aggregated_mask,
    )
    # Blend original image with the one, where instances are colored.
    # As a result instances masks become transparent.
    cv2.addWeighted(frame, 0.5, segments_image, 0.5, 0, dst=frame)

    if put_object_count:
        put_text_on_rect_bg(frame, f"Obj. count: {len(predictions)}", (0, 0))
    return frame


def put_labels(frame: Mat, predictions: List[Annotation]):
    """Converts predictions to text labels and puts them to the top left corner of a frame."""

    frame = frame.copy()
    assert len(predictions) == 1
    # TODO (ilya-krylov): handle multi-label classification
    assert len(predictions[0].get_labels()) == 1
    label = predictions[0].get_labels()[0]
    color = tuple(getattr(label.color, x) for x in ("blue", "green", "red"))
    put_text_on_rect_bg(frame, label.name, (0, 0), color=color)
    return frame


def draw_bounding_boxes(frame: Mat, predictions: List[Annotation], put_object_count: bool):
    """Converts predictions to bounding boxes and draws them on a frame."""

    frame = frame.copy()
    height, width = frame.shape[0], frame.shape[1]
    for prediction in predictions:
        if isinstance(prediction.shape, Rectangle):
            x1 = int(prediction.shape.x1 * width)
            x2 = int(prediction.shape.x2 * width)
            y1 = int(prediction.shape.y1 * height)
            y2 = int(prediction.shape.y2 * height)
            assert len(prediction.get_labels()) == 1
            label = prediction.get_labels()[0]
            color = tuple(getattr(label.color, x) for x in ("blue", "green", "red"))
            cv2.rectangle(frame, (x1, y1), (x2, y2), color, thickness=2)
            put_text_on_rect_bg(frame, label.name, (x1, y1), color=color)
        else:
            warn(
                f"Predictions called on Annotations with shape {type(prediction.shape)}."
                "Expected shape to be of type Rectangle."
            )

    if put_object_count:
        put_text_on_rect_bg(frame, f"Obj. count: {len(predictions)}", (0, 0))
    return frame


def draw_predictions(task_type: TaskType, predictions: List[Annotation], frame: Mat, fit_to_size: Tuple[int, int]):
    """Converts predictions to visual representations depending on task type and draws them on a frame."""

    width, height = frame.shape[1], frame.shape[0]
    if fit_to_size:
        ratio_x = fit_to_size[0] / width
        ratio_y = fit_to_size[1] / height
        ratio = min(ratio_x, ratio_y)
        frame = cv2.resize(frame, None, fx=ratio, fy=ratio)
    if task_type in {TaskType.DETECTION, TaskType.ANOMALY_DETECTION}:
        frame = draw_bounding_boxes(frame, predictions, put_object_count=True)
    elif task_type in {TaskType.CLASSIFICATION, TaskType.ANOMALY_CLASSIFICATION}:
        frame = put_labels(frame, predictions)
    elif task_type in {TaskType.INSTANCE_SEGMENTATION, TaskType.ROTATED_DETECTION}:
        frame = draw_masks(frame, predictions, put_object_count=True)
    elif task_type in {TaskType.SEGMENTATION, TaskType.ANOMALY_SEGMENTATION}:
        frame = draw_masks(frame, predictions, put_object_count=False)
    else:
        raise ValueError(f"Unknown task type: {task_type}")
    return frame



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


import logging
import os
from collections import namedtuple
from copy import deepcopy
from typing import TYPE_CHECKING, List, Type

if TYPE_CHECKING:
    from nncf.torch.nncf_network import NNCFNetwork

from tasks import NNCFTask

from otx.algorithms.anomaly.adapters.anomalib.data.mvtec import OtxMvtecDataset
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model_template import TaskType
from tests.test_suite.training_tests_actions import (
    BaseOTXTestAction,
    OTXTestExportAction,
    OTXTestExportEvaluationAction,
    OTXTestNNCFAction,
    OTXTestNNCFEvaluationAction,
    OTXTestNNCFExportAction,
    OTXTestNNCFExportEvaluationAction,
    OTXTestNNCFGraphAction,
    OTXTestNNCFValidationFQAction,
    OTXTestPotAction,
    OTXTestPotEvaluationAction,
    OTXTestPotValidationFQAction,
    OTXTestTrainingAction,
    OTXTestTrainingEvaluationAction,
)
from tests.test_suite.training_tests_common import ROOT_PATH_KEY, make_paths_be_abs

logger = logging.getLogger(__name__)


def DATASET_PARAMETERS_FIELDS() -> List[str]:
    return deepcopy(["dataset_path"])


DatasetParameters = namedtuple("DatasetParameters", DATASET_PARAMETERS_FIELDS())  # type: ignore


def _get_dataset_params_from_dataset_definitions(dataset_definitions, dataset_name):
    if dataset_name not in dataset_definitions:
        raise ValueError(
            f"dataset {dataset_name} is absent in dataset_definitions, "
            f"dataset_definitions.keys={list(dataset_definitions.keys())}"
        )
    cur_dataset_definition = dataset_definitions[dataset_name]
    training_parameters_fields = {k: v for k, v in cur_dataset_definition.items() if k in DATASET_PARAMETERS_FIELDS()}
    print(f"training_parameters_fields: {training_parameters_fields}")
    make_paths_be_abs(training_parameters_fields, dataset_definitions[ROOT_PATH_KEY])
    print(f"training_parameters_fields after make_paths_be_abs: {training_parameters_fields}")

    assert set(DATASET_PARAMETERS_FIELDS()) == set(
        training_parameters_fields.keys()
    ), f"ERROR: dataset definitions for name={dataset_name} does not contain all required fields"
    assert all(
        training_parameters_fields.values()
    ), f"ERROR: dataset definitions for name={dataset_name} contains empty values for some required fields"

    params = DatasetParameters(**training_parameters_fields)
    return params


def _create_anomaly_dataset_and_labels_schema(
    dataset_params: DatasetParameters, dataset_name: str, task_type: TaskType
):
    logger.debug(f"Path to dataset: {dataset_params.dataset_path}")
    category_list = [f.path for f in os.scandir(dataset_params.dataset_path) if f.is_dir()]
    items = []
    if "short" in dataset_name:
        logger.debug(f"Creating short dataset {dataset_name}")
        items.extend(OtxMvtecDataset(path=dataset_params.dataset_path, seed=0, task_type=task_type).generate())
    else:
        for category in category_list:
            logger.debug(f"Creating dataset for {category}")
            items.extend(OtxMvtecDataset(path=category, seed=0, task_type=task_type).generate())
    dataset = DatasetEntity(items=items)
    labels = dataset.get_labels()
    labels_schema = LabelSchemaEntity.from_labels(labels)
    return dataset, labels_schema


def get_anomaly_domain_test_action_classes(
    anomaly_domain_test_train_action: OTXTestTrainingAction,
) -> List[Type[BaseOTXTestAction]]:
    return [
        anomaly_domain_test_train_action,
        OTXTestTrainingEvaluationAction,
        OTXTestExportAction,
        OTXTestExportEvaluationAction,
        OTXTestPotAction,
        OTXTestPotEvaluationAction,
        OTXTestPotValidationFQAction,
        OTXTestNNCFAction,
        OTXTestNNCFEvaluationAction,
        OTXTestNNCFExportAction,
        OTXTestNNCFExportEvaluationAction,
        OTXTestNNCFGraphAction,
        OTXTestNNCFValidationFQAction,
    ]


def get_dummy_compressed_model(task: NNCFTask) -> "NNCFNetwork":
    """
    Return compressed model without initialization
    """
    from anomalib.utils.callbacks.nncf.utils import wrap_nncf_model

    # Disable quantaizers initialization
    for compression in task.optimization_config["nncf_config"]["compression"]:
        if compression["algorithm"] == "quantization":
            compression["initializer"] = {"batchnorm_adaptation": {"num_bn_adaptation_samples": 0}}

    _, compressed_model = wrap_nncf_model(task.model, task.optimization_config["nncf_config"])
    return compressed_model



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from abc import ABCMeta, abstractmethod


class BaseScalarScheduler(metaclass=ABCMeta):
    def __init__(self):
        super(BaseScalarScheduler, self).__init__()

    def __call__(self, step, epoch_size) -> float:
        return self._get_value(step, epoch_size)

    @abstractmethod
    def _get_value(self, step, epoch_size) -> float:
        raise NotImplementedError("Subclass must implement this method")
